import unittest

import numpy as np

from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import full_range
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.basetest_scattering import HelperScattering
from osaft.tests.solution_factory import SolutionFactory


class TestScattering(BaseTestSolutions, HelperScattering):
    def setUp(self) -> None:
        super().setUp()

        self._v_boundary_conditions = 1e-3
        self._v_fluid_threshold = 1e-8
        self.runs = 5

        self.cls = SolutionFactory().doinikov_2021_viscoelastic_scattering()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Test V_r, V_theta
    # -------------------------------------------------------------------------

    def test_V(self) -> None:
        combinations = [
            (True, False),
            (False, True),
            (True, True),
        ]
        threshold = 1e-8
        for factor in np.linspace(1, 5, self.n_runs):
            for n in range(0, 5):
                for scattered, incident in combinations:
                    r = factor * self.R_0
                    self.assertAlmostEqual(
                        self.V_r(n, r, scattered, incident),
                        self.cls.V_r(n, r, scattered, incident),
                        threshold=threshold,
                    )
                    self.assertAlmostEqual(
                        self.d_V_r(n, r, scattered, incident),
                        self.cls.d_V_r(n, r, scattered, incident),
                        threshold=threshold,
                    )
                    self.assertAlmostEqual(
                        self.d2_V_r(n, r, scattered, incident),
                        self.cls.d2_V_r(n, r, scattered, incident),
                        threshold=threshold,
                    )
                    self.assertAlmostEqual(
                        self.V_theta(n, r, scattered, incident),
                        self.cls.V_theta(n, r, scattered, incident),
                    )
                    self.assertAlmostEqual(
                        self.d_V_theta(n, r, scattered, incident),
                        self.cls.d_V_theta(n, r, scattered, incident),
                        threshold=threshold,
                    )
                    self.assertAlmostEqual(
                        self.d2_V_theta(n, r, scattered, incident),
                        self.cls.d2_V_theta(n, r, scattered, incident),
                        threshold=threshold,
                    )
                    self.assertAlmostEqual(
                        self.phi(n, r, scattered, incident),
                        self.cls.phi(n, r, scattered, incident),
                        threshold=threshold,
                    )
                    self.assertAlmostEqual(
                        self.d_phi(n, r, scattered, incident),
                        self.cls.d_phi(n, r, scattered, incident),
                        threshold=threshold,
                    )

    def test_V_error(self):
        self.assertRaises(ValueError, self.cls.V_r, 1, self.R_0, False, False)
        self.assertRaises(
            ValueError,
            self.cls.d_V_r,
            1,
            self.R_0,
            False,
            False,
        )
        self.assertRaises(
            ValueError,
            self.cls.d2_V_r,
            1,
            self.R_0,
            False,
            False,
        )
        self.assertRaises(
            ValueError,
            self.cls.V_theta,
            1,
            self.R_0,
            False,
            False,
        )
        self.assertRaises(
            ValueError,
            self.cls.d_V_theta,
            1,
            self.R_0,
            False,
            False,
        )
        self.assertRaises(
            ValueError,
            self.cls.d2_V_theta,
            1,
            self.R_0,
            False,
            False,
        )
        self.assertRaises(ValueError, self.cls.phi, 1, self.R_0, False, False)
        self.assertRaises(
            ValueError,
            self.cls.d_phi,
            1,
            self.R_0,
            False,
            False,
        )

    # -------------------------------------------------------------------------
    # Velocity Methods
    # -------------------------------------------------------------------------

    def radial_acoustic_fluid_velocity(
        self,
        r,
        theta,
        t,
        scattered,
        incident,
        mode=None,
    ) -> complex:
        def coef(l: int, x: float) -> complex:
            return self.cls.V_r(l, x, scattered, incident)

        out = np.exp(-1j * self.omega * t)
        if mode is not None:
            out *= Leg.cos_monomial(mode, theta, coef(mode, r))
        else:
            out *= Leg.cos_poly(
                theta,
                np.array(
                    [coef(n, r) for n in full_range(0, self.N_max)],
                ),
            )
        return out

    def tangential_acoustic_fluid_velocity(
        self,
        r,
        theta,
        t,
        scattered,
        incident,
        mode=None,
    ) -> complex:
        def coef(l: int, x: float) -> complex:
            return self.cls.V_theta(l, x, scattered, incident)

        out = np.exp(-1j * self.omega * t)
        if mode is not None:
            out *= Leg.first_cos_monomial(
                mode,
                theta,
                coef(mode, r),
            )
        else:
            out *= Leg.first_cos_poly(
                theta,
                np.array(
                    [coef(n, r) for n in full_range(0, self.N_max)],
                ),
            )
        return out

    def radial_particle_displacement(self, r, theta, t, mode) -> complex:
        def coef(l: int, x: float) -> complex:
            first = self.k_l * self.a_hat(l) * Bf.d1_besselj(l, self.k_l * x)
            second = l * (l + 1) * self.b_hat(l) * Bf.besselj(l, self.k_t * x)
            second /= x
            return first - second

        out = np.exp(-1j * self.omega * t)
        if mode is not None:
            out *= Leg.cos_monomial(mode, theta, coef(mode, r))
        else:
            out *= Leg.cos_poly(
                theta,
                np.array(
                    [coef(n, r) for n in full_range(0, self.N_max)],
                ),
            )
        return out

    def radial_particle_velocity(self, r, theta, t, mode) -> complex:
        return (-1j * self.omega) * self.radial_particle_displacement(
            r,
            theta,
            t,
            mode,
        )

    def tangential_particle_displacement(self, r, theta, t, mode) -> complex:
        def coef(l: int, x: float) -> complex:
            first = self.a_hat(l) * Bf.besselj(l, self.k_l * x)
            second = Bf.besselj(l, self.k_t * x)
            second += self.k_t * x * Bf.d1_besselj(l, self.k_t * x)
            second *= self.b_hat(l)
            return (first - second) / x

        out = np.exp(-1j * self.omega * t)
        if mode is not None:
            out *= Leg.first_cos_monomial(
                mode,
                theta,
                coef(mode, r),
            )
        else:
            out *= Leg.first_cos_poly(
                theta,
                np.array(
                    [coef(n, r) for n in full_range(1, self.N_max)],
                ),
            )
        return out

    def tangential_particle_velocity(self, r, theta, t, mode) -> complex:
        return (-1j * self.omega) * self.tangential_particle_displacement(
            r,
            theta,
            t,
            mode,
        )

    # -------------------------------------------------------------------------
    # Helper Methods
    # ------------------------------------------------------------------------

    @property
    def N_max(self) -> int:
        return self.cls.N_max

    @property
    def omega(self) -> float:
        return self.f * 2 * np.pi

    @property
    def k_f(self) -> complex:
        return self.cls.k_f

    @property
    def k_v(self) -> complex:
        return self.cls.k_v

    @property
    def k_l(self) -> complex:
        return self.cls.k_l

    @property
    def k_t(self) -> complex:
        return self.cls.k_t

    def a(self, n: int) -> complex:
        return self.cls.a(n)

    def a_hat(self, n: int) -> complex:
        return self.cls.a_hat(n)

    def b(self, n: int) -> complex:
        return self.cls.b(n)

    def b_hat(self, n: int) -> complex:
        return self.cls.b_hat(n)

    def A_in(self, n: int) -> complex:
        return self.cls.A_in(n)

    # -------------------------------------------------------------------------
    # V_r, V_theta methods
    # ------------------------------------------------------------------------

    def V_r(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            return (
                self.a(n) * self.k_f * Bf.d1_hankelh1(n, self.k_f * r)
                - n * (n + 1) * self.b(n) * Bf.hankelh1(n, self.k_v * r) / r
            )
        elif incident and not scattered:
            return self.A_in(n) * self.k_f * Bf.d1_besselj(n, self.k_f * r)
        elif incident and scattered:
            return self.V_r(n, r, True, False) + self.V_r(n, r, False, True)
        else:
            return 0

    def d_V_r(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            return (
                self.k_f**2 * self.a(n) * Bf.d2_hankelh1(n, self.k_f * r)
                - n
                * (n + 1)
                * self.b(n)
                * (
                    self.k_v * Bf.d1_hankelh1(n, self.k_v * r) * r
                    - Bf.hankelh1(n, self.k_v * r)
                )
                / r**2
            )
        elif incident and not scattered:
            return (
                self.k_f**2 * self.A_in(n) * Bf.d2_besselj(n, self.k_f * r)
            )
        elif incident and scattered:
            return self.d_V_r(n, r, True, False) + self.d_V_r(
                n,
                r,
                False,
                True,
            )
        else:
            return 0

    def d2_V_r(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            return self.k_f**3 * self.a(n) * Bf.d3_hankelh1(
                n,
                self.k_f * r,
            ) - n * (n + 1) * self.b(n) * (
                self.k_v**2 * Bf.d2_hankelh1(n, self.k_v * r) / r
                + 2
                * (
                    Bf.hankelh1(n, self.k_v * r)
                    - r * self.k_v * Bf.d1_hankelh1(n, self.k_v * r)
                )
                / r**3
            )
        elif not scattered and incident:
            return (
                self.k_f**3 * self.A_in(n) * Bf.d3_besselj(n, self.k_f * r)
            )
        elif scattered and incident:
            return self.d2_V_r(n, r, True, False) + self.d2_V_r(
                n,
                r,
                False,
                True,
            )
        else:
            return 0

    def V_theta(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            return (
                self.a(n) * Bf.hankelh1(n, self.k_f * r)
                - self.b(n)
                * (
                    Bf.hankelh1(n, self.k_v * r)
                    + self.k_v * r * Bf.d1_hankelh1(n, self.k_v * r)
                )
            ) / r
        elif not scattered and incident:
            return self.A_in(n) * Bf.besselj(n, self.k_f * r) / r
        elif scattered and incident:
            if n == 0:
                return 0
            return self.V_theta(n, r, True, False) + self.V_theta(
                n,
                r,
                False,
                True,
            )
        else:
            return 0

    def d_V_theta(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            nominator_div_r = self.V_theta(n, r, scattered, incident)
            d_nominator = self.k_f * self.a(n) * Bf.d1_hankelh1(
                n,
                self.k_f * r,
            ) - self.b(n) * self.k_v * (
                2 * Bf.d1_hankelh1(n, self.k_v * r)
                + self.k_v * r * Bf.d2_hankelh1(n, self.k_v * r)
            )
            return (d_nominator - nominator_div_r) / r
        elif incident and not scattered:
            nominator_div_r = self.V_theta(n, r, scattered, incident)
            d_nominator = (
                self.k_f * self.A_in(n) * Bf.d1_besselj(n, self.k_f * r)
            )
            return (d_nominator - nominator_div_r) / r
        elif scattered and incident:
            if n == 0:
                return 0
            return self.d_V_theta(n, r, True, False) + self.d_V_theta(
                n,
                r,
                False,
                True,
            )
        else:
            return 0

    def d2_V_theta(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            dd_nominator = self.k_f**2 * self.a(n) * Bf.d2_hankelh1(
                n,
                self.k_f * r,
            ) - self.b(n) * self.k_v**2 * (
                3 * Bf.d2_hankelh1(n, self.k_v * r)
                + self.k_v * r * Bf.d3_hankelh1(n, self.k_v * r)
            )
            return (
                dd_nominator - 2 * self.d_V_theta(n, r, scattered, incident)
            ) / r
        elif incident and not scattered:
            dd_nominator = (
                self.k_f**2 * self.A_in(n) * Bf.d2_besselj(n, self.k_f * r)
            )
            return (
                dd_nominator - 2 * self.d_V_theta(n, r, scattered, incident)
            ) / r
        elif scattered and incident:
            if n == 0:
                return 0
            return self.d2_V_theta(n, r, True, False) + self.d2_V_theta(
                n,
                r,
                False,
                True,
            )

    def phi(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            return self.a(n) * Bf.hankelh1(n, self.k_f * r)
        elif incident and not scattered:
            return self.A_in(n) * Bf.besselj(n, self.k_f * r)
        elif scattered and incident:
            return self.phi(n, r, True, False) + self.phi(n, r, False, True)
        else:
            return 0

    def d_phi(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        if scattered and not incident:
            return self.a(n) * self.k_f * Bf.d1_hankelh1(n, self.k_f * r)
        elif not scattered and incident:
            return self.A_in(n) * self.k_f * Bf.d1_besselj(n, self.k_f * r)
        elif scattered and incident:
            return self.d_phi(n, r, True, False) + self.d_phi(
                n,
                r,
                False,
                True,
            )
        else:
            return 0


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from osaft import ViscoelasticFluid, WaveType
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.solutions.basedoinikov2021.scattering import (
    BaseScatteringDoinikov2021,
)


class ScatteringField(BaseScatteringDoinikov2021):
    """Scattering field class for Doinikov (viscous fluid-elastic sphere; 2021)

    :param f: frequency [Hz]
    :param R_0: radius [m]
    :param rho_s: density of the particle [kg/m^3]
    :param E_s: Young's modulus of the particle [Pa]
    :param nu_s: Poisson's ratio of the particle [-]
    :param rho_f: density of the fluid [kg/m^3]
    :param c_f: speed of sound in the fluid [m/s]
    :param eta_f: fluid component shear viscosity [Pa s]
    :param zeta_f: fluid component bulk viscosity [Pa s]
    :param eta_p: shear viscosity of polymer component [Pa s]
    :param zeta_p: bulk viscosity of polymer component [Pa s]
    :param lambda_M: relaxation time of fluid [s]
    :param p_0: pressure amplitude [Pa]
    :param wave_type: wave type
    :param position: position in the standing wave field [m]
    """

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float,
        rho_s: float,
        E_s: float,
        nu_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        eta_p: float,
        zeta_p: float,
        lambda_M: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
        N_max: int = 5,
    ) -> None:

        # Fluid attribute (note: the frequency will be set in the base class
        # only to make sure all attributes have the same Frequency instance)
        fluid = ViscoelasticFluid(
            frequency=f,
            rho=rho_f,
            c=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            eta_p=eta_p,
            zeta_p=zeta_p,
            lambda_M=lambda_M,
        )

        # Init of parent class
        super().__init__(
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            E_s=E_s,
            nu_s=nu_s,
            fluid=fluid,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
            N_max=N_max,
        )

    # -------------------------------------------------------------------------
    # Required Methods
    # -------------------------------------------------------------------------

    @property
    def eta(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscoelasticFluid.eta_c`"""
        return self.fluid.eta_c

    @property
    def zeta(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscoelasticFluid.zeta_c`"""
        return self.fluid.zeta_c

    # -------------------------------------------------------------------------
    # Wrappers specific to viscoelastic model
    # -------------------------------------------------------------------------

    @property
    def eta_p(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscoelasticFluid.eta_p`"""
        return self.fluid.eta_p

    @eta_p.setter
    def eta_p(self, value) -> None:
        self.fluid.eta_p = value

    @property
    def zeta_p(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscoelasticFluid.zeta_p`"""
        return self.fluid.zeta_p

    @zeta_p.setter
    def zeta_p(self, value) -> None:
        self.fluid.zeta_p = value

    @property
    def lambda_M(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscoelasticFluid.lambda_M`"""
        return self.fluid.lambda_M

    @lambda_M.setter
    def lambda_M(self, value) -> None:
        self.fluid.lambda_M = value


if __name__ == "__main__":
    pass



from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from typing import Any

import numpy as np

from osaft import log
from osaft.core.functions import full_range


class BaseVariable(ABC):
    """BaseVariable class defining basic methods and attributes shared between
    active and passive attributes.

    This class and its children are based on the Observer Design Pattern.
    In the present context value changes shall be tracked so
    :class:`ActiveVariable` know when their dependencies changed, and they
    need to recompute their underlying value.

    :param name: Name of the variable
    :raises: TypeError if :attr:`name` is of wrong type
    """

    def __init__(self, name: None | str = None) -> None:
        """Constructor method"""
        self._value = None
        self._variables_to_notify: list[BaseVariable] = []

        if name is None:
            self._name = "Unnamed"
        elif isinstance(name, str):
            self._name = name
        else:
            raise TypeError(f"name is of type {type(name)} instead of str")

    @property
    def name(self) -> str:
        """Returns name of the variable.

        :setter: sets :attr:`~osaft.core.variable.BaseVariable._name`
        """
        return self._name

    @property
    def value(self) -> Any:
        """Value of the attribute

        :setter: just defined in :class:`PassiveVariable`
        :getter: returns/computes the value of the instance
        """
        return self._value

    def notify(self) -> None:
        """Notifies all items in :attr:`BaseVariable._variables_to_notify` that
        a dependent value changed. Just the information of a changed dependency
        is passed.
        """
        log.debug(f"Variable {repr(self)} notifies:")
        for att in self._variables_to_notify:
            log.debug(f"attribute: {att}")
            att._change_update_status()

    def _link(self, input_variable: Any) -> None:
        """Links the inputs to self, so it knows who needs to update when
        :attr:`BaseVariable.value` changes.

        :param input_variable:
        """
        log.info(
            f"Variables linked: {str(input_variable)} depends on"
            f" {str(self)}",
        )
        self._variables_to_notify.append(input_variable)

    @staticmethod
    def _check_types(
        inputs: tuple,
    ) -> None:
        """Checks if elements of :attr:`inputs` are of type
        :class:`BaseVariable`

        :param inputs: iterable to be checked
        :raises: TypeError if :attr:`inputs` is of wrong type
        """
        for item in inputs:
            if not isinstance(item, BaseVariable):
                out = f"{item} is of type {type(item)} instead of BaseVariable"
                raise TypeError(out)


class PassiveVariable(BaseVariable):
    """Passive variable without any computation dependencies

    :param val: initial value
    :param name: Name of the variable
    """

    def __init__(
        self,
        val: Any,
        name: None | str = None,
    ) -> None:
        """Constructor method"""
        super().__init__(name)
        self.value = val

        # no if clause needed because last in inheritance
        log.debug(f"Creating {self}")

    @BaseVariable.value.setter
    def value(self, value: Any) -> None:
        """
        Checks if :attr:`BaseVariable._value` is already equal to value.
        If not, first :attr:`BaseVariable._value` is set to `value`
        and then all item in :attr:`BaseVariable._variables_to_notify` are
        notified.

        :param value: new value
        """
        has_changed = True
        if self._value is not None:
            try:
                L1 = len(self._value)
            except TypeError:
                L1 = 0
            try:
                L2 = len(value)
            except TypeError:
                L2 = 0

            if L1 == 0 and L2 == 0:
                has_changed = self._value != value
            elif L1 == L2:
                has_changed = not np.alltrue(self._value == value)

        if has_changed:
            log.debug(f"value change of {repr(self)} to {value}")
            self._value = value
            self.notify()

    def __repr__(self) -> str:
        return f"PassiveVariable(name={self.name}, value={self.value})"

    def __str__(self):
        return f"{self.name} = {self.value}"


class BaseActiveVariable(BaseVariable, ABC):
    """Base for variables that depended on at least one :class:`.BaseVariable`

    :param function_call: function signature that defines how
        :attr:`ActiveVariable._value` is computed
        Callable[[], Any]
    :param name: Name of the variable
    """

    def __init__(
        self,
        function_call: Callable[[], Any],
        name: None | str = None,
    ) -> None:
        """Constructor method"""
        super().__init__(name)
        self._update_function = function_call
        self._needs_update = True

        # no if clause needed because last in inheritance
        log.debug(f"Creating {self}")

    @property
    def needs_update(self) -> bool:
        """Returns if an update is needed for the next value retrieval"""
        return self._needs_update

    def _change_update_status(self) -> None:
        """Set :attr:`ActiveVariable._needs_update` to `True` and calls
        :meth:`BaseVariable.notify`.
        """
        log.debug(f"{repr(self)} changes _needs_update to True")
        self._needs_update = True
        self.notify()

    def is_computed_by(self, *variables: Any) -> None:
        """Adds the notification dependencies

        :param variables: variables that are referenced in
            :meth:`~ActiveVariable._update`
        :raises: TypeError if no arguments are passed
        """
        if len(variables) == 0:
            out = "is_computed_by() takes at least one positional argument "
            out += "(0 were given)"
            raise TypeError(out)
        self._check_types(variables)
        for variable in variables:
            variable._link(self)

    def __repr__(self) -> str:
        if not self._needs_update:
            return f"ActiveVariable(name={self.name}, value={self.value})"
        else:
            return f"ActiveVariable(name={self.name}, value=N.A.)"

    def __str__(self):
        if not self._needs_update:
            return f"{self.name} = {self.value}"
        else:
            return f"{self.name} = N.A."


class ActiveVariable(BaseActiveVariable):
    """Variables that depended on at least one :class:`.BaseVariable`

    :param function_call: function signature that defines how
        :attr:`ActiveVariable._value` is computed
    :param name: Name of the variable
    """

    def __init__(
        self,
        function_call: Callable[[], Any],
        name: None | str = None,
    ) -> None:
        """Constructor method"""
        super().__init__(function_call, name)

    @property
    def value(self) -> Any:
        """Returns the value of the attribute. If
        :attr:`ActiveVariable.needs_update` is true, it calls
        :meth:`ActiveVariable._update` first and then changes
        :attr:`ActiveVariable._needs_update` to `False` before returning the
        value.
        """
        if self.needs_update:
            self._update()
        return self._value

    def _update(self) -> None:
        """Executes the assigned routine and outputs its value."""
        log.debug(f"{repr(self)} recomputes its value")
        self._value = self._update_function()
        self._needs_update = False
        log.debug(f"Recomputed: {repr(self)}")


class ActiveListVariable(BaseActiveVariable):
    """ActiveVariables with an expendable list as :attr:`value`

    :param function_call: function signature that defines how the n-the
    element of the `list` :attr:`ActiveVariable._value` is computed
    :param name: Name of the variable
    """

    def __init__(
        self,
        function_call: Callable[[int], Any],
        name: None | str = None,
    ) -> None:
        """Constructor method"""
        super().__init__(function_call, name)

    @property
    def value(self) -> list:
        """Returns the `list` containing the data of the instance. If
        :attr:`ActiveVariable.needs_update` is true, it calls the list is
        emptied, :attr:`ActiveVariable._needs_update` to `False` and an
        empty list is returned.
        """
        if self._needs_update:
            self._reset_value()
        return self._value

    def item(self, index: int) -> Any:
        """Returns the item stored at `item` index. If the value variable
        needs recomputation first, the whole `list` up to the index is
        recomputed. If the index does not exist then all the missing items
        up to `index` are computed.

        :param index: index
        """
        if index < len(self.value):
            return self._value[index]
        else:
            self._update(index)
            return self._value[index]

    def _update(self, index: int) -> None:
        """Computes the missing values in the list up to `index`.

        :param index: index
        """
        before = len(self._value)
        for n in full_range(before, index):
            self._value.append(self._update_function(n))
        log.debug(f"Computed: {repr(self)} up to index = {index}")
        self._needs_update = False

    def _reset_value(self):
        """Resets the value to an empty list."""
        log.debug(f"{repr(self)} recomputes its value")
        self._value = []


if __name__ == "__main__":
    pass



import unittest

from osaft.core.functions import pi, sin
from osaft.tests.basetest_arf import (
    HelperCompareARF,
    HelperStandingARF,
    HelperTravelingARF,
)
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestGorkov(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self.cls = SolutionFactory().gorkov_1962_arf()
        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    @property
    def f_1(self):
        return 1 - self.cls.kappa_s / self.cls.kappa_f

    @property
    def f_2(self):
        rho_tilde = self.rho_s / self.rho_f
        return 2 * (rho_tilde - 1) / (2 * rho_tilde + 1)

    @property
    def Phi(self):
        return self.f_1 / 3 + self.f_2 / 2

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_properties(self) -> None:
        properties = ["f_1", "f_2", "Phi"]
        self._test_properties(properties)

    def test_not_small_particle(self):
        # This method is not actually testing, only for coverage
        self.cls.R_0 = self.cls.fluid.lambda_f
        self.cls.compute_arf()


class TestStandingARF(TestGorkov, HelperStandingARF):
    def compute_arf(self):
        out = 4 * pi * self.cls.Phi * self.cls.k_f * self.R_0**3
        out *= self.cls.E_ac * sin(2 * self.position)
        return out


class TestTraveling(TestGorkov, HelperTravelingARF):
    def compute_arf(self):
        A = self.p_0 / (1j * 2 * pi * self.f * self.rho_f)
        out = 2 / 9 * pi * self.rho_f * abs(A) ** 2
        out *= (self.cls.k_f * self.R_0) ** 6
        out *= self.f_1**2 + self.f_1 * self.f_2 + 3 / 4 * self.f_2**2
        return out


class TestCompareToKing(TestGorkov, HelperCompareARF):
    def setUp(self) -> None:

        super().setUp()

        self.parameters.R_0 = 1e-7
        self.parameters._R_0.high = 1e-6

        self.parameters.c_s = 1e10
        self.parameters._c_s.low = 1e9

        self.arf_compare_threshold = 1e-4

        self.cls = SolutionFactory().gorkov_1962_arf()
        self.compare_cls = SolutionFactory().king_1934_arf()
        self.compare_cls.small_particle_limit = True

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == "__main__":
    unittest.main()



import unittest
from abc import abstractmethod
from collections.abc import Callable
from random import randint, random

import numpy as np

from osaft import ViscoelasticFluid, ViscousFluid, WaveType
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import exp, full_range, pi
from osaft.tests.testing_parameters import RandomParameters


class BaseHelperScattering:
    @staticmethod
    def get_random_n():
        return randint(0, 10)

    @staticmethod
    def get_random_r(cls):
        return cls.R_0 * (2 + 20 * random())

    @staticmethod
    def get_random_t():
        return random()

    @staticmethod
    def get_random_theta():
        return random() * pi


class HelperScattering(BaseHelperScattering):

    parameters: RandomParameters

    # -------------------------------------------------------------------------
    # Thresholds
    # -------------------------------------------------------------------------

    _v_particle_threshold = 1e-12
    _v_fluid_threshold = 1e-12
    _v_boundary_conditions = 1e-12

    @property
    def v_particle_threshold(self) -> float:
        return self._v_particle_threshold

    @property
    def v_fluid_threshold(self) -> float:
        return self._v_fluid_threshold

    @property
    def v_boundary_conditions(self) -> float:
        return self._v_boundary_conditions

    # -------------------------------------------------------------------------
    # Acoustic Velocities Methods
    # -------------------------------------------------------------------------

    def V_r_i(self, n, r) -> complex:
        out = self.cls.A_in(n)
        out *= self.cls.k_f
        out *= Bf.d1_besselj(n, self.cls.k_f * r)
        return out

    def V_theta_i(self, n, r) -> complex:
        out = self.cls.A_in(n)
        out *= Bf.besselj(n, self.cls.k_f * r)
        out /= r
        return out

    def V_r_sc(self, n, r) -> complex:
        pass

    def V_theta_sc(self, n, r) -> complex:
        pass

    def radial_acoustic_fluid_velocity(
        self,
        r,
        theta,
        t,
        scattered,
        incident,
        mode=None,
    ) -> float:
        if scattered and incident:

            def v(l, x):
                return self.V_r_sc(l, x) + self.V_r_i(l, x)

        elif scattered and not incident:
            v = self.V_r_sc
        elif not scattered and incident:
            v = self.V_r_i
        else:

            def v(l, x):
                return 0

        out = exp(-1j * self.cls.omega * t)
        if mode is not None:
            out *= Leg.cos_monomial(mode, theta, v(mode, r))
        else:
            out *= Leg.cos_poly(
                theta,
                [v(n, r) for n in full_range(self.cls.N_max)],
            )
        return out

    def tangential_acoustic_fluid_velocity(
        self,
        r,
        theta,
        t,
        scattered,
        incident,
        mode=None,
    ) -> float:
        if scattered and incident:

            def v(l, x):
                if l == 0:
                    return 0
                return self.V_theta_sc(l, x) + self.V_theta_i(l, x)

        elif scattered and not incident:
            v = self.V_theta_sc
        elif not scattered and incident:
            v = self.V_theta_i
        else:

            def v(l, x):
                return 0

        out = exp(-1j * self.cls.omega * t)
        if mode is not None:
            out *= Leg.first_cos_monomial(mode, theta, v(mode, r))
        else:
            out *= Leg.first_cos_poly(
                theta,
                [v(n, r) for n in full_range(0, self.cls.N_max)],
            )
        return out

    def radial_particle_displacement(self, r, theta, t, mode):
        velocity = self.radial_particle_velocity(r, theta, t, mode)
        return velocity / (-1j * self.cls.omega)

    def tangential_particle_displacement(self, r, theta, t, mode):
        velocity = self.tangential_particle_velocity(r, theta, t, mode)
        return velocity / (-1j * self.cls.omega)

    @abstractmethod
    def radial_particle_velocity(self, r, theta, t, mode) -> float:
        pass

    @abstractmethod
    def tangential_particle_velocity(self, r, theta, t, mode) -> float:
        pass

    # -------------------------------------------------------------------------
    # Acoustic Velocities Tests
    # -------------------------------------------------------------------------

    def test_radial_particle_velocity(self) -> None:
        # skip changing R_0 so the RT.is_inside_sphere does not get triggered
        self.parameters.list_parameters.remove(self.parameters._R_0)
        for n in range(self.n_runs):
            n = self.get_random_n()
            r = self.cls.R_0 * 0.9
            t = self.get_random_t()
            theta = self.get_random_theta()
            self.do_testing(
                func_1=self.radial_particle_velocity,
                args_1=(r, theta, t, n),
                func_2=self.cls.radial_particle_velocity,
                args_2=(r, theta, t, n),
                threshold=self.v_particle_threshold,
            )

    def test_tangential_particle_velocity(self) -> None:
        # skip changing R_0 so the RT.is_inside_sphere does not get triggered
        self.parameters.list_parameters.remove(self.parameters._R_0)
        for _ in range(self.n_runs):
            n = self.get_random_n()
            r = self.cls.R_0 * 0.9
            t = self.get_random_t()
            theta = self.get_random_theta()
            self.do_testing(
                func_1=self.tangential_particle_velocity,
                args_1=(r, theta, t, n),
                func_2=self.cls.tangential_particle_velocity,
                args_2=(r, theta, t, n),
                threshold=self.v_particle_threshold,
            )

    def test_radial_particle_displacement(self) -> None:
        # skip changing R_0 so the RT.is_inside_sphere does not get triggered
        self.parameters.list_parameters.remove(self.parameters._R_0)
        for n in range(self.n_runs):
            n = self.get_random_n()
            r = self.cls.R_0 * 0.9
            t = self.get_random_t()
            theta = self.get_random_theta()
            self.do_testing(
                func_1=self.radial_particle_displacement,
                args_1=(r, theta, t, n),
                func_2=self.cls.radial_particle_displacement,
                args_2=(r, theta, t, n),
                threshold=self.v_particle_threshold,
            )

    def test_tangential_particle_displacement(self) -> None:
        # skip changing R_0 so the RT.is_inside_sphere does not get triggered
        self.parameters.list_parameters.remove(self.parameters._R_0)
        for _ in range(self.n_runs):
            n = self.get_random_n()
            r = self.cls.R_0 * 0.9
            t = self.get_random_t()
            theta = self.get_random_theta()
            self.do_testing(
                func_1=self.tangential_particle_displacement,
                args_1=(r, theta, t, n),
                func_2=self.cls.tangential_particle_displacement,
                args_2=(r, theta, t, n),
                threshold=self.v_particle_threshold,
            )

    def test_fluid_velocities_Error(self) -> None:
        # neither scattering nor incident is selected
        with self.assertRaises(ValueError):
            self.cls.radial_acoustic_fluid_velocity(
                self.cls.R_0,
                0,
                0,
                False,
                False,
            )
        with self.assertRaises(ValueError):
            self.cls.tangential_acoustic_fluid_velocity(
                self.cls.R_0,
                0,
                0,
                False,
                False,
            )

    def test_radial_acoustic_fluid_velocity(self) -> None:
        # skip changing R_0 so the RT.is_outside_sphere does not get triggered
        self.parameters.list_parameters.remove(self.parameters._R_0)
        for incident, scattering in [
            (True, True),
            (True, False),
            (False, True),
        ]:
            for _ in range(5):
                n = self.get_random_n()
                r = self.get_random_r(self.cls)
                t = self.get_random_t()
                theta = self.get_random_theta()

                self.do_testing(
                    func_1=self.radial_acoustic_fluid_velocity,
                    args_1=(r, t, theta, incident, scattering, n),
                    func_2=self.cls.radial_acoustic_fluid_velocity,
                    args_2=(r, t, theta, incident, scattering, n),
                    threshold=self.v_fluid_threshold,
                )
                # test once with mode = None
                self.do_testing(
                    func_1=self.radial_acoustic_fluid_velocity,
                    args_1=(r, t, theta, incident, scattering),
                    func_2=self.cls.radial_acoustic_fluid_velocity,
                    args_2=(r, t, theta, incident, scattering),
                    threshold=self.v_fluid_threshold,
                )

    def test_tangential_acoustic_fluid_velocity(self) -> None:
        # skip changing R_0 so the RT.is_outside_sphere does not get triggered
        self.parameters.list_parameters.remove(self.parameters._R_0)
        for incident, scattering in [
            (True, True),
            (True, False),
            (False, True),
        ]:
            for _ in range(5):
                n = self.get_random_n()
                r = self.get_random_r(self.cls)
                t = self.get_random_t()
                theta = self.get_random_theta()

                self.do_testing(
                    func_1=self.tangential_acoustic_fluid_velocity,
                    args_1=(r, theta, t, incident, scattering, n),
                    func_2=self.cls.tangential_acoustic_fluid_velocity,
                    args_2=(r, theta, t, incident, scattering, n),
                    threshold=self.v_fluid_threshold,
                )

                # test once with mode = None
                self.do_testing(
                    func_1=self.tangential_acoustic_fluid_velocity,
                    args_1=(r, t, theta, incident, scattering),
                    func_2=self.cls.tangential_acoustic_fluid_velocity,
                    args_2=(r, t, theta, incident, scattering),
                    threshold=self.v_fluid_threshold,
                )

    # -------------------------------------------------------------------------
    # Test Input Testers
    # -------------------------------------------------------------------------

    def test_ThetaTester_present(self):
        lst = [1.0, 0.5, 4]

        self.assertRaises(
            ValueError,
            self.cls.radial_particle_velocity,
            0,
            lst,
            0,
        )
        self.assertRaises(
            ValueError,
            self.cls.tangential_particle_velocity,
            0,
            lst,
            0,
        )

        self.assertRaises(
            ValueError,
            self.cls.radial_acoustic_fluid_velocity,
            0,
            lst,
            0,
            True,
            True,
        )
        self.assertRaises(
            ValueError,
            self.cls.tangential_acoustic_fluid_velocity,
            0,
            lst,
            0,
            True,
            True,
        )

    def test_RadiusTester_present(self):
        lst = 0.99 * self.R_0 * np.random.rand(10)
        lst[0] = -1.0
        lst[-1] = 1.1 * self.R_0

        self.assertRaises(
            ValueError,
            self.cls.radial_particle_velocity,
            lst,
            0,
            0,
        )
        self.assertRaises(
            ValueError,
            self.cls.tangential_particle_velocity,
            lst,
            0,
            0,
        )

        lst = 1.01 * self.R_0 * (np.random.rand(10) + 1)
        lst[0] = -1.0
        lst[-1] = 0.1 * self.R_0

        self.assertRaises(
            ValueError,
            self.cls.radial_acoustic_fluid_velocity,
            lst,
            0,
            0,
            True,
            True,
        )
        self.assertRaises(
            ValueError,
            self.cls.tangential_acoustic_fluid_velocity,
            lst,
            0,
            0,
            True,
            True,
        )

    # -------------------------------------------------------------------------
    # Test Boundary Conditions
    # -------------------------------------------------------------------------

    def test_boundary_conditions(self) -> None:
        test_tangential = False
        if isinstance(self.cls.fluid, ViscousFluid) or isinstance(
            self.cls.fluid,
            ViscoelasticFluid,
        ):
            test_tangential = True

        modes_to_test = [0, 1, 2, 3, 4, None]

        thetas = np.random.uniform(size=10)
        # testing radial direction at r=R_0
        for theta in thetas:
            t = self.get_random_t()
            for mode in modes_to_test:
                with self.subTest(
                    msg="Testing radial condition",
                    theta=f"\u03C0 * {theta}",
                    t=t,
                    mode=mode,
                ):
                    v_s = self.cls.radial_acoustic_fluid_velocity(
                        self.R_0,
                        pi * theta,
                        t,
                        True,
                        True,
                        mode=mode,
                    )
                    v_p = self.cls.radial_particle_velocity(
                        self.R_0,
                        pi * theta,
                        t,
                        mode=mode,
                    )
                    self.assertAlmostEqual(
                        v_p,
                        v_s,
                        threshold=self.v_boundary_conditions,
                    )

        if test_tangential:
            # testing tangential direction at r=R_0
            for theta in thetas:
                t = self.get_random_t()
                for mode in modes_to_test:
                    with self.subTest(
                        msg="Testing tangential condition",
                        theta=f"\u03C0 * {theta}",
                        t=t,
                        mode=mode,
                    ):
                        v_s = self.cls.tangential_acoustic_fluid_velocity(
                            self.R_0,
                            pi * theta,
                            t,
                            True,
                            True,
                        )
                        v_p = self.cls.tangential_particle_velocity(
                            self.R_0,
                            pi * theta,
                            t,
                        )
                        self.assertAlmostEqual(
                            v_p,
                            v_s,
                            threshold=self.v_boundary_conditions,
                        )


class HelperCompareScattering(BaseHelperScattering):
    _scattering_compare_threshold = 1e-12

    @property
    def scattering_compare_threshold(self) -> float:
        return self._scattering_compare_threshold

    def _sc_comparison(
        self,
        func_1: Callable,
        func_2: Callable,
        msg: str,
        inside_sphere: bool,
    ) -> None:

        if inside_sphere:
            test_R = self.R_0 * np.random.uniform(low=0, high=1, size=10)
        else:
            test_R = self.R_0 * np.random.uniform(low=1, high=10, size=10)

        test_theta = np.pi * np.random.uniform(low=0, high=1, size=10)

        t = self.get_random_t()
        for wt in WaveType:
            self.cls.wave_type = wt
            self.compare_cls.wave_type = wt
            for r in np.nditer(test_R):
                for theta in np.nditer(test_theta):
                    if inside_sphere:
                        v1 = func_1(r, theta, t)
                        v2 = func_2(r, theta, t)
                    else:
                        v1 = func_1(r, theta, t, True, True)
                        v2 = func_2(r, theta, t, True, True)

                    subtest_msg = (
                        f"{msg}, "
                        f"theta={theta/np.pi:1.2f}*\u03C0, "
                        f"r={r/self.R_0:1.2f}*R_0 , "
                        f"t={t:1.2e}"
                    )

                    with self.subTest(msg=subtest_msg):
                        self.assertAlmostEqual(
                            v1,
                            v2,
                            threshold=self.scattering_compare_threshold,
                        )

    @unittest.skip("unstable test")
    def test_sc_comparison_outside(self) -> None:
        msg = "Tangential fluid velocity"
        func_1 = self.cls.tangential_acoustic_fluid_velocity
        func_2 = self.compare_cls.tangential_acoustic_fluid_velocity
        self._sc_comparison(func_1, func_2, msg, inside_sphere=False)

        msg = "Radial fluid velocity"
        func_1 = self.cls.radial_acoustic_fluid_velocity
        func_2 = self.compare_cls.radial_acoustic_fluid_velocity
        self._sc_comparison(func_1, func_2, msg, inside_sphere=False)

    @unittest.skip("unstable test")
    def test_sc_comparison_inside(self) -> None:
        msg = "Tangential particle velocity"
        func_1 = self.cls.tangential_particle_velocity
        func_2 = self.compare_cls.tangential_particle_velocity
        self._sc_comparison(func_1, func_2, msg, inside_sphere=True)

        msg = "Radial particle velocity"
        func_1 = self.cls.radial_particle_velocity
        func_2 = self.compare_cls.radial_particle_velocity
        self._sc_comparison(func_1, func_2, msg, inside_sphere=True)



from __future__ import annotations

import numpy as np

from osaft import log
from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import exp
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveListVariable
from osaft.solutions.base_scattering import (
    BaseScattering,
    BaseScatteringRigidParticle,
)
from osaft.solutions.king1934.base import BaseKing

NDArray = np.ndarray


class ScatteringField(BaseKing, BaseScatteringRigidParticle, BaseScattering):
    """Scattering field class for King (1934)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of incident wave (traveling/standing)
    :param position: Position in the standing wave field [rad]
    :param N_max: Highest order mode included in the computation [-]
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        rho_f: float,
        c_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method"""

        # init of parent class
        BaseKing.__init__(
            self,
            f,
            R_0,
            rho_s,
            rho_f,
            c_f,
            p_0,
            wave_type,
            position,
        )
        BaseScattering.__init__(self, N_max)

        # Dependent variables
        self._A_dash_n = ActiveListVariable(self._compute_A_dash_n, "A_dash_n")
        self._A_dash_n.is_computed_by(
            self.field._A_in,
            self._alpha,
            self._rho_tilde,
        )
        if type(self) is ScatteringField:
            log.info(str(self))
            log.debug(repr(self))

    def __repr__(self):
        return (
            f"{type(self)}(f={self.f}, R_0={self.R_0}, "
            f"rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, "
            f"p_0={self.p_0}, position={self.position}, "
            f"wave_type={self.wave_type}, N_max={self.N_max})"
        )

    def __str__(self):
        out = "King's solution to the Scattering field"
        out += " with following properties: \n"
        out += SF.get_str_text("Wave type", "", self.wave_type, "")
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "rad",
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_0",
            self.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_f",
            self.kappa_f,
            "1/Pa",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        out += "Other\n"
        out += SF.get_str_text(
            "#modes",
            "N_max",
            self.N_max,
            None,
        )
        return out

    # -----------------------------------------------------
    # User-facing Methods
    # -----------------------------------------------------

    def particle_velocity(self, t: float) -> float:
        """Particle velocity

        Returns the value of the particle velocity
        in the direction of the axis of rotational
        symmetry of the radiation field in [m/s]

        :param t: time [s]
        """

        out = exp(-1j * self.omega * t) * self.A_in(1) / self.alpha**3
        out *= self.k_f / self.rho_tilde
        out /= self.F_n(1, self.alpha) - 1j * self.G_n(1, self.alpha)
        return out

    # -----------------------------------------------------
    # Coefficients
    # -----------------------------------------------------

    def potential_coefficient(self, n: int) -> complex:
        return -self.A_dash_n(n)

    @staticmethod
    def phi_n(n: int, arg: float) -> complex:
        r"""From King equation (22)

        Returns :math:`\phi_n`

        :param n: mode
        :param arg: argument for spherical Bessel function
        """
        return -Bf.bessely(n, arg) * arg ** (-n)

    @staticmethod
    def psi_n(n: int, arg: float) -> complex:
        r"""From King equation (22)

        Returns :math:`\psi_n`

        :param n: mode
        :param arg: argument for spherical Bessel function
        """
        return Bf.besselj(n, arg) * arg ** (-n)

    def F_n(self, n: int, arg: float) -> float:
        """From King equation (30) and (31)

        Returns :math:`F_n`

        :param n: mode
        :param arg: argument for spherical Bessel function
        """
        out = self.alpha**2 * self.phi_n(n + 1, arg)
        if n == 1:
            out -= (1 - 1 / self.rho_tilde) * self.phi_n(n, arg)
        else:
            out -= n * self.phi_n(n, arg)

        return out

    def G_n(self, n: int, arg: float) -> complex:
        """From King equation (30) and (31)

        Returns :math:`G_n`

        :param n: mode
        :param arg: argument for spherical Bessel function
        """
        out = self.alpha**2 * self.psi_n(n + 1, arg)
        if n == 1:
            out -= (1 - 1 / self.rho_tilde) * self.psi_n(n, arg)
        else:
            out -= n * self.psi_n(n, arg)

        return out

    def A_dash_n(self, n: int) -> complex:
        r"""From King equation (21), (28)
        potential from 28 to form of 21

        Returns Scattering field coefficient :math:`A^{\prime}_{n}`

        :param n: mode number
        """
        return self._A_dash_n.item(n)

    def _compute_A_dash_n(self, n: int) -> complex:
        out = -self.A_in(n) * self.G_n(n, self.alpha)
        out /= self.G_n(n, self.alpha) + 1j * self.F_n(n, self.alpha)
        return out

    # -----------------------------------------------------
    # Methods
    # -----------------------------------------------------

    def Phi_scattering(
        self,
        r: float | NDArray | list[float],
        theta: float | NDArray | list[float],
        t: float | NDArray | list[float],
    ) -> complex:
        r"""King equation (28) for scattering potential
        :math:`\Phi_{\mathrm{scattering}}`

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        """

        def Phi_n(n: int):
            num = -self.G_n(n, self.alpha)
            num *= self.A_in(n) * Bf.hankelh2(n, r * self.k_f)
            denom = self.G_n(n, self.alpha)
            denom += 1j * self.F_n(n, self.alpha)
            coefficient = num / denom

            out = exp(-1j * self.omega * t)
            out *= Leg.cos_monomial(n, theta, coefficient)
            return out

        out = 0
        for n in range(self.N_max + 1):
            out += Phi_n(n)
        return out

    def Phi_incident(
        self,
        r: float | NDArray | list[float],
        theta: float | NDArray | list[float],
        t: float | NDArray | list[float],
    ) -> complex:
        r"""King equation (28) for scattering potential
        :math:`\Phi_{\mathrm{incident}}`

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        """

        def Phi_n(n: int):
            coefficient = self.A_in(n) * Bf.besselj(n, r * self.k_f)
            out = exp(-1j * self.omega * t)
            out *= Leg.cos_monomial(n, theta, coefficient)
            return out

        out = 0
        for n in range(self.N_max + 1):
            out += Phi_n(n)
        return out

    def V_r_sc(
        self,
        n: int,
        r: float | NDArray | list[float],
    ) -> complex:
        """Implements :meth:`osaft.solutions.base_scattering.BaseScattering.\
V_r_sc`
        """
        return self.A_dash_n(n) * self.k_f * Bf.d1_hankelh2(n, self.k_f * r)

    def V_theta_sc(
        self,
        n: int,
        r: float | NDArray | list[float],
    ) -> complex:
        """Implements :meth:`osaft.solutions.base_scattering.BaseScattering.\
V_theta_sc`
        """
        return self.A_dash_n(n) * Bf.hankelh2(n, self.k_f * r) / r


if __name__ == "__main__":
    pass



from __future__ import annotations

import osaft
from osaft import log
from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveListVariable
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.hasegawa1969.scattering import ScatteringField


class ARF(ScatteringField, BaseARF):
    """ARF class for Hasegawa & Yosioka (1969)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the solid sphere [kg/m^3]
    :param E_s: Young's modulus [Pa]
    :param nu_s: Poisson's ratio of the solid sphere [-]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param position: Position in the standing wave field [rad]
    :param wave_type: Either standing or progressive wave
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        E_s: float,
        nu_s: float,
        rho_f: float,
        c_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method"""

        ScatteringField.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            E_s=E_s,
            nu_s=nu_s,
            rho_f=rho_f,
            c_f=c_f,
            p_0=p_0,
            position=position,
            wave_type=wave_type,
            N_max=N_max,
        )

        self._U_n = ActiveListVariable(
            self._compute_U_n,
            "Coefficient U_n",
        )
        self._V_n = ActiveListVariable(
            self._compute_V_n,
            "Coefficient V_n",
        )
        self._d_U_n = ActiveListVariable(
            self._compute_d_U_n,
            "Derivative of Coefficient U_n",
        )
        self._V_n_d1 = ActiveListVariable(
            self._compute_d_V_n,
            "Derivative of Coefficient V_n",
        )

        self._V_n.is_computed_by(
            self.sphere._R_0,
            self._c_n,
            self.fluid._k_f,
        )
        self._U_n.is_computed_by(
            self.sphere._R_0,
            self._c_n,
            self.fluid._k_f,
        )
        self._V_n_d1.is_computed_by(
            self.sphere._R_0,
            self._c_n,
            self.fluid._k_f,
        )
        self._d_U_n.is_computed_by(
            self.sphere._R_0,
            self._c_n,
            self.fluid._k_f,
        )

        log.info(str(self))
        log.debug(repr(self))

    def __repr__(self):
        return (
            f"{type(self)}(f={self.f}, R_0={self.R_0}, "
            f"rho_s={self.rho_s}, rho_f={self.rho_f}, "
            f"c_f={self.c_f}, E={self.E_s}, nu = {self.nu_s}, "
            f"p_0={self.p_0}, position={self.position}, "
            f"wave_type={self.wave_type}, N_max={self.N_max}"
        )

    def __str__(self):
        out = "Hasegawa's solution to the ARF field"
        out += " with following properties: \n"
        out += SF.get_str_text("Wave type", "", self.wave_type, "")
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "rad",
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_f",
            self.c_f,
            "m/s",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Young's Modulus",
            "E",
            self.E_s,
            "Pa",
        )

        out += "Other\n"
        out += SF.get_str_text(
            "#modes",
            "N_max",
            self.N_max,
            None,
        )
        return out

    # -------------------------------------------------------------------------
    # User-faced functions
    # -------------------------------------------------------------------------

    def compute_arf(self) -> float:
        """
        Acoustic radiation force [N]

        Computes the ARF, based on the general solution (Eq. 13)

        :raises WrongWaveTypeError: if wrong :attr:`wave_type`
        """

        # Checking wave_type
        self.check_wave_type()
        if self.wave_type == WaveType.STANDING:
            return self._compute_arf_standing()
        else:
            return self._compute_arf_travelling()

    # -------------------------------------------------------------------------
    # General Solution
    # -------------------------------------------------------------------------

    def _compute_arf_travelling(self) -> float:
        """
        Computes general analytical solution for the ARF in the propagating
        wave case

        (Eq. 13)
        """

        out = 5 * [0]

        for n in range(self.N_max):
            # (Eq. 23)
            s1 = self.d_U_n(n) * self.d_V_n(n + 1)
            s1 -= self.d_V_n(n) * self.d_U_n(n + 1)
            s1 *= n + 1
            out[0] += s1

            # (Eq. 24)
            s2 = self.U_n(n) * self.V_n(n + 1)
            s2 -= self.V_n(n) * self.U_n(n + 1)
            s2 *= n * (n + 1) * (n + 2)
            out[1] += s2

            # (Eq. 25)
            s3 = self.U_n(n) * self.d_V_n(n + 1)
            s3 -= self.V_n(n) * self.d_U_n(n + 1)
            s3 *= n * (n + 1)
            s4 = self.d_U_n(n) * self.V_n(n + 1)
            s4 -= self.d_V_n(n) * self.U_n(n + 1)
            s4 *= (n + 1) * (n + 2)
            out[2] += s3
            out[3] += s4

            # (Eq. 26)
            s5 = self.U_n(n) * self.V_n(n + 1)
            s5 -= self.V_n(n) * self.U_n(n + 1)
            s5 *= n + 1
            out[4] += s5

        out[0] *= -self.x_f**2
        out[1] *= 1
        out[2] *= self.x_f
        out[3] *= -self.x_f
        out[4] *= -self.x_f**2

        # (Eq. 13)
        out = 2 * osaft.pi * self.rho_f * sum(out) * abs(self.field.A) ** 2
        return out

    def _compute_arf_standing(self) -> float:
        """
        Computes general analytical solution for the ARF. This is from the
        Hasegawa1979 paper.

        (Eq. 10)
        """

        out = 5 * [0]

        for n in range(self.N_max):
            # (Eq. 15, rows 1 and 2)
            s1 = self.d_U_n(n) * self.d_U_n(n + 1)
            s1 += self.d_V_n(n) * self.d_V_n(n + 1)
            s1 *= (n + 1) * (-1) ** (n + 1)
            out[0] += s1
            # check

            # (Eq. 15, rows 3 and 4)
            s2 = self.U_n(n) * self.U_n(n + 1)
            s2 += self.V_n(n) * self.V_n(n + 1)
            s2 *= n * (n + 1) * (n + 2) * (-1) ** (n + 1)
            out[1] += s2

            # (Eq. 15, rows 5 and 6)
            s3 = self.U_n(n) * self.d_U_n(n + 1)
            s3 += self.V_n(n) * self.d_V_n(n + 1)
            s3 *= n * (n + 1) * (-1) ** (n + 1)
            # (Eq. 15, row 6)
            s4 = self.d_U_n(n) * self.U_n(n + 1)
            s4 += self.d_V_n(n) * self.V_n(n + 1)
            s4 *= (n + 1) * (n + 2) * (-1) ** (n + 1)
            out[2] += s3
            out[3] += s4

            # (Eq. 15, row 7)
            s5 = self.U_n(n) * self.U_n(n + 1)
            s5 += self.V_n(n) * self.V_n(n + 1)
            s5 *= (n + 1) * (-1) ** (n + 1)
            out[4] += s5

        common_coeff = (
            1
            * osaft.pi
            * self.rho_f
            * self.field.A**2
            * osaft.core.functions.sin(
                2 * self.position / self.k_f.real * self.k_f,
            )
        )

        out[0] *= -self.x_f**2 * common_coeff
        out[1] *= common_coeff
        out[2] *= self.x_f * common_coeff
        out[3] *= -self.x_f * common_coeff
        out[4] *= -self.x_f**2 * common_coeff

        # (Eq. 13)
        out = -sum(out)

        return out.real

    # -------------------------------------------------------------------------
    # Coefficients
    # -------------------------------------------------------------------------

    def U_n(self, n: int) -> complex:
        """Coefficient :math:`U_n` [-]

        (Eq. 19)

        :param n: mode [-]
        :return: coefficient [m^2/s]
        """
        return self._U_n.item(n)

    def V_n(self, n: int) -> complex:
        """Coefficient :math:`V_n` [-]

        (Eq. 19)

        :param n: mode [-]
        :return: coefficient [m^2/s]
        """
        return self._V_n.item(n)

    def d_U_n(self, n: int) -> complex:
        """Coefficient :math:`U_n'` [-]

        (Eq. 19)

        :param n: mode [-]
        :return: coefficient [m^2/s]
        """
        return self._d_U_n.item(n)

    def d_V_n(self, n: int) -> complex:
        """Coefficient :math:`V_n'` [-]

        (Eq. 19)

        :param n: mode [-]
        :return: coefficient [m^2/s]
        """
        return self._V_n_d1.item(n)

    def _compute_U_n(self, n: int) -> complex:
        """
        Compute U_n according to Eq. (19)
        """
        c_n = self.c_n(n)
        out = (1 + c_n.real) * Bf.besselj(
            n,
            self.k_f * self.R_0,
        ) + c_n.imag * Bf.bessely(n, self.k_f * self.R_0)
        return out

    def _compute_V_n(self, n: int) -> complex:
        """
        Compute V_n according to Eq. (19)
        """
        c_n = self.c_n(n)
        out = c_n.imag * Bf.besselj(
            n,
            self.k_f * self.R_0,
        ) - c_n.real * Bf.bessely(n, self.k_f * self.R_0)
        return out

    def _compute_d_U_n(self, n: int) -> complex:
        """
        Compute first derivative of U_n according to Eq. (19)
        """
        c_n = self.c_n(n)
        out = (1 + c_n.real) * Bf.d1_besselj(
            n,
            self.k_f * self.R_0,
        ) + c_n.imag * Bf.d1_bessely(n, self.k_f * self.R_0)
        return out

    def _compute_d_V_n(self, n: int) -> complex:
        """
        Compute first derivative of V_n according to Eq. (19)
        """
        c_n = self.c_n(n)
        out = c_n.imag * Bf.d1_besselj(
            n,
            self.k_f * self.R_0,
        ) - c_n.real * Bf.d1_bessely(n, self.k_f * self.R_0)
        return out


if __name__ == "__main__":
    pass



from __future__ import annotations

from abc import ABC
from collections.abc import Iterable, Sequence

import numpy as np

from osaft import ViscoelasticFluid, ViscousFluid, WaveType, log
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import full_range
from osaft.core.geometries import Sphere
from osaft.core.helper import InputHandler
from osaft.core.variable import ActiveListVariable, ActiveVariable
from osaft.solutions.base_scattering import BaseScattering
from osaft.solutions.basedoinikov2021.coefficientmatrix import (
    CoefficientMatrix,
)

NDArray = np.ndarray


class BaseScatteringDoinikov2021(CoefficientMatrix, BaseScattering, ABC):
    """Scattering field class for Doinikov (viscous fluid-elastic sphere; 2021)

    :param f: frequency [Hz]
    :param R_0: radius [m]
    :param rho_s: density of the particle [kg/m^3]
    :param E_s: Young's modulus of the particle [Pa]
    :param nu_s: Poisson's ratio of the particle [-]
    :param fluid: Fluid instance of the model
    :param p_0: pressure amplitude [Pa]
    :param wave_type: wave type
    :param position: position in the standing wave field [m]
    """

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float,
        rho_s: float,
        E_s: float,
        nu_s: float,
        fluid: ViscousFluid | ViscoelasticFluid,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method"""

        BaseScattering.__init__(self, N_max=N_max)

        CoefficientMatrix.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            E_s=E_s,
            nu_s=nu_s,
            fluid=fluid,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
        )

        log.info("Create ScatteringField")

        # Dependent variables - Coefficients
        self._a = ActiveListVariable(
            self._compute_a,
            "radial fluid coefficients a_n",
        )
        self._a.is_computed_by(self._list_matrix_M_n, self._list_vector_n)
        self._a_hat = ActiveListVariable(
            self._compute_a_hat,
            "radial particle coefficients a_hat_n",
        )
        self._a_hat.is_computed_by(self._list_matrix_M_n, self._list_vector_n)
        self._b = ActiveListVariable(
            self._compute_b,
            "tangential fluid coefficients b_n",
        )
        self._b.is_computed_by(self._list_matrix_M_n, self._list_vector_n)
        self._b_hat = ActiveListVariable(
            self._compute_b_hat,
            "tangential particle coefficients " "b_hat_n",
        )
        self._b_hat.is_computed_by(self._list_matrix_M_n, self._list_vector_n)

        # Dependent Variables - Ranges
        self._range_N_max = ActiveVariable(self._compute_range_N_max)
        self._range_1_N_max = ActiveVariable(self._compute_range_1_N_max)
        self._range_2_N_max = ActiveVariable(self._compute_range_2_N_max)

    def __repr__(self):
        return (
            f"ScatteringField({self.f}, {self.R_0}, {self.rho_s}, "
            f"{self.E_s}), {self.nu_s}. {self.rho_f}, {self.c_f}, "
            f"{self.eta_f}, {self.zeta_f}, {self.p_0}, {self.wave_type}, "
            f"{self.field.position}, {self.N_max})"
        )

    # -------------------------------------------------------------------------
    # User-facing methods
    # -------------------------------------------------------------------------

    def radial_particle_displacement(
        self,
        r: float | Sequence,
        theta: float | Sequence,
        t: float | Sequence,
        mode: None | int = None,
    ) -> complex | NDArray:
        """First-order radial particle displacements [m]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        :return: first-order radial particle displacement
        """

        def coefficients(
            l: int,
            x: complex | NDArray,
        ) -> complex | NDArray:
            term1 = self.k_l * self.a_hat(l) * Bf.d1_besselj(l, self.k_l * x)
            term2 = (l + 1) * l * self.b_hat(l) * Bf.besselj(l, self.k_t * x)
            term2 /= x
            return term1 - term2

        r, theta, t = InputHandler.handle_input(
            r,
            theta,
            t,
            self.R_0,
            inside_sphere=True,
        )

        return self.radial_mode_superposition(coefficients, r, theta, t, mode)

    def radial_particle_velocity(
        self,
        r: float | Sequence,
        theta: float | Sequence,
        t: float | Sequence,
        mode: None | int = None,
    ) -> complex | NDArray:
        """First-order radial particle velocity [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        displacement = self.radial_particle_displacement(r, theta, t, mode)
        return -1j * self.frequency.omega * displacement

    def tangential_particle_displacement(
        self,
        r: float | Sequence,
        theta: float | Sequence,
        t: float | Sequence,
        mode: None | int = None,
    ) -> complex | NDArray:
        """First-order tangential particle displacements [m]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        :return: first-order tangential particle displacements
        """

        def coefficients(
            l: int,
            x: complex | NDArray,
        ) -> complex | NDArray:
            term1 = Bf.besselj(l, self.k_l * x)
            term2 = Bf.besselj(l, self.k_t * x)
            term2 += self.k_t * x * Bf.d1_besselj(l, self.k_t * x)
            return (self.a_hat(l) * term1 - self.b_hat(l) * term2) / x

        r, theta, t = InputHandler.handle_input(
            r,
            theta,
            t,
            self.R_0,
            inside_sphere=True,
        )

        return self.tangential_mode_superposition(
            coefficients,
            r,
            theta,
            t,
            mode,
        )

    def tangential_particle_velocity(
        self,
        r: float | Sequence,
        theta: float | Sequence,
        t: float | Sequence,
        mode: None | int = None,
    ) -> complex | NDArray:
        """First-order tangential particle displacements [m]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        :return: first-order tangential particle displacements
        """

        displacement = self.tangential_particle_displacement(r, theta, t, mode)
        return -1j * self.frequency.omega * displacement

    def potential_coefficient(self, n: int) -> complex:
        return self.a(n)

    # -------------------------------------------------------------------------
    # Coefficients
    # -------------------------------------------------------------------------

    def _compute_a(self, n: int) -> complex:
        if n == 0:
            return self.a_0()
        else:
            return self.det_M_n(n, 0) / self.det_M_n(n)

    def _compute_a_hat(self, n: int) -> complex:
        if n == 0:
            return self.a_hat_0()
        else:
            return self.det_M_n(n, 2) / self.det_M_n(n)

    def _compute_b(self, n: int) -> complex:
        if n == 0:
            return 0
        else:
            return self.det_M_n(n, 1) / self.det_M_n(n)

    def _compute_b_hat(self, n: int) -> complex:
        if n == 0:
            return 0
        else:
            return self.det_M_n(n, 3) / self.det_M_n(n)

    def a(self, n: int) -> complex:
        """Coefficient :math:`a_n` [m^2/s]

        (Eq. A18)

        :param n: mode
        :return: coefficient a_n
        """
        return self._a.item(n)

    def a_hat(self, n: int) -> complex:
        """Coefficient :math:`\\hat{a}_n` [m^2/s]

        (Eq. A18)

        :param n: mode
        :return: coefficient a_hat_n
        """
        return self._a_hat.item(n)

    def b(self, n: int) -> complex:
        """Coefficient :math:`b_n` [m^2/s]

        (Eq. A18)

        :param n: mode
        """
        return self._b.item(n)

    def b_hat(self, n: int) -> complex:
        """Coefficient :math:`\\hat{b}_n` [m^2/s]

        (Eq. A18)

        :param n: mode
        """
        return self._b_hat.item(n)

    # -------------------------------------------------------------------------
    # Ranges
    # -------------------------------------------------------------------------

    @property
    def range_N_max(self) -> Iterable:
        """Returns ``range(0, N_max + 1)``"""
        return self._range_N_max.value

    @property
    def range_1_N_max(self) -> Iterable:
        """Returns ``range(1, N_max + 1)``"""
        return self._range_1_N_max.value

    @property
    def range_2_N_max(self) -> Iterable:
        """Returns ``range(2, N_max + 1)``"""
        return self._range_2_N_max.value

    def _compute_range_N_max(self) -> Iterable:
        return full_range(self.N_max)

    def _compute_range_1_N_max(self) -> Iterable:
        return full_range(1, self.N_max)

    def _compute_range_2_N_max(self) -> Iterable:
        return full_range(2, self.N_max)

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def V_r_sc(self, n: int, r: float) -> complex:
        """Scattering contribution to :math:`V_{rn}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        """
        inviscid = self.a(n) * self.k_f * Bf.d1_hankelh1(n, self.k_f * r)
        viscous = n * (n + 1) * self.b(n) * Bf.hankelh1(n, self.k_v * r) / r
        return inviscid - viscous

    def d_V_r(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        """:math:`\\partial_r V_{rn}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        :param scattered: scattered field
        :param incident: incident field
        """
        if incident and not scattered:
            return self.d_V_r_i(n, r)
        elif incident and scattered:
            return self.d_V_r_sc(n, r) + self.d_V_r_i(n, r)
        elif not incident and scattered:
            return self.d_V_r_sc(n, r)
        else:
            raise ValueError(
                "Neither scattered nor incident field has "
                "been has been selected. Velocity field is zero.",
            )

    def d_V_r_sc(self, n: int, r: float) -> complex:
        """:math:`\\partial_r V_{rn}^{sc}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        """
        viscous = self.k_v * Bf.d1_hankelh1(n, self.k_v * r) * r
        viscous -= Bf.hankelh1(n, self.k_v * r)
        viscous *= n * (n + 1) * self.b(n) / r**2
        inviscid = self.k_f**2 * self.a(n) * Bf.d2_hankelh1(n, self.k_f * r)
        return inviscid - viscous

    def d_V_r_i(self, n: int, r: float) -> complex:
        """:math:`\\partial_r V_{rn}^{i}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.k_f**2 * self.A_in(n) * Bf.d2_besselj(n, self.k_f * r)

    def d2_V_r(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        """:math:`\\partial^2_r V_{rn}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        :param scattered: scattered field
        :param incident: incident field
        """
        if incident and not scattered:
            return self.d2_V_r_i(n, r)
        elif incident and scattered:
            return self.d2_V_r_sc(n, r) + self.d2_V_r_i(n, r)
        elif not incident and scattered:
            return self.d2_V_r_sc(n, r)
        else:
            raise ValueError(
                "Neither scattered nor incident field has "
                "been has been selected. Velocity field is zero.",
            )

    def d2_V_r_i(self, n: int, r: float) -> complex:
        """:math:`\\partial^2_r V_{rn}^{i}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.k_f**3 * self.A_in(n) * Bf.d3_besselj(n, self.k_f * r)

    def d2_V_r_sc(self, n: int, r: float) -> complex:
        """:math:`\\partial^2_r V_{rn}^{sc}(r)` [m/s]

        (Eq. 35)

        :param n: mode
        :param r: radial coordinate [m]
        """
        inviscid = self.k_f**3 * self.a(n) * Bf.d3_hankelh1(n, self.k_f * r)
        viscous = 2 * Bf.hankelh1(n, self.k_v * r) / r**3
        viscous -= 2 * self.k_v * Bf.d1_hankelh1(n, self.k_v * r) / r**2
        viscous += self.k_v**2 * Bf.d2_hankelh1(n, self.k_v * r) / r
        viscous *= n * (n + 1) * self.b(n)
        return inviscid - viscous

    def V_theta_sc(self, n: int, r: float) -> complex:
        """Scattering contribution to :math:`V_{\\theta n}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        """
        inviscid = self.a(n) * Bf.hankelh1(n, self.k_f * r)
        viscous = Bf.hankelh1(n, self.k_v * r)
        viscous += self.k_v * r * Bf.d1_hankelh1(n, self.k_v * r)
        viscous *= self.b(n)
        return (inviscid - viscous) / r

    def d_V_theta(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        """:math:`\\partial_r V_{\\theta n}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        :param scattered: scattered field
        :param incident: incident field
        """
        if incident and scattered:
            if n == 0:
                return 0
            return self.d_V_theta_sc(n, r) + self.d_V_theta_i(n, r)
        elif incident and not scattered:
            return self.d_V_theta_i(n, r)
        elif not incident and scattered:
            return self.d_V_theta_sc(n, r)
        else:
            raise ValueError(
                "Neither scattered nor incident field has "
                "been has been selected. Velocity field is zero.",
            )

    def d_V_theta_sc(
        self,
        n: int,
        r: float,
    ) -> complex:
        """:math:`\\partial_r V_{\\theta n}^{sc}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        """
        nominator_div_r = self.V_theta_sc(n, r)
        inviscid = self.k_f * self.a(n) * Bf.d1_hankelh1(n, self.k_f * r)
        viscous = 2 * Bf.d1_hankelh1(n, self.k_v * r)
        viscous += self.k_v * r * Bf.d2_hankelh1(n, self.k_v * r)
        viscous *= self.b(n) * self.k_v
        d_nominator = inviscid - viscous
        return (d_nominator - nominator_div_r) / r

    def d_V_theta_i(self, n: int, r: float) -> complex:
        """:math:`\\partial_r V_{\\theta n}^{i}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        """
        nominator_div_r = self.V_theta_i(n, r)
        d_nominator = self.k_f * self.A_in(n) * Bf.d1_besselj(n, self.k_f * r)
        return (d_nominator - nominator_div_r) / r

    def d2_V_theta(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        """:math:`\\partial^2_r V_{\\theta n}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        :param scattered: scattered field
        :param incident: incident field
        """
        if incident and scattered:
            if n == 0:
                return 0
            return self.d2_V_theta_sc(n, r) + self.d2_V_theta_i(n, r)
        elif incident and not scattered:
            return self.d2_V_theta_i(n, r)
        elif not incident and scattered:
            return self.d2_V_theta_sc(n, r)
        else:
            raise ValueError(
                "Neither scattered nor incident field has "
                "been has been selected. Velocity field is zero.",
            )

    def d2_V_theta_sc(self, n: int, r: float) -> complex:
        """:math:`\\partial^2_r V_{\\theta n}^{sc}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        """
        inviscid = self.k_f**2 * self.a(n) * Bf.d2_hankelh1(n, self.k_f * r)
        viscous = 3 * Bf.d2_hankelh1(n, self.k_v * r)
        viscous += self.k_v * r * Bf.d3_hankelh1(n, self.k_v * r)
        viscous *= self.b(n) * self.k_v**2
        dd_nominator = inviscid - viscous
        return (dd_nominator - 2 * self.d_V_theta_sc(n, r)) / r

    def d2_V_theta_i(self, n: int, r: float) -> complex:
        """:math:`\\partial^2_r V_{\\theta n}^{i}(r)` [m/s]

        (Eq. 36)

        :param n: mode
        :param r: radial coordinate [m]
        """
        dd_nominator = self.A_in(n) * Bf.d2_besselj(n, self.k_f * r)
        dd_nominator *= self.k_f**2
        return (dd_nominator - 2 * self.d_V_theta_i(n, r)) / r

    def phi(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        """:math:`\\varphi_n(r)`

        (Eq. 28, 30)

        :param n: mode
        :param r: radial coordinate [m]
        :param scattered: scattered field
        :param incident: incident field
        """
        if incident and not scattered:
            return self.phi_i(n, r)
        elif incident and scattered:
            return self.phi_i(n, r) + self.phi_sc(n, r)
        elif not incident and scattered:
            return self.phi_sc(n, r)
        else:
            raise ValueError(
                "Neither scattered nor incident field has "
                "been has been selected. Velocity field is zero.",
            )

    def phi_sc(self, n: int, r: float) -> complex:
        """:math:`\\varphi_n^{sc}(r)`

        (Eq. 28, 30)

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.a(n) * Bf.hankelh1(n, self.k_f * r)

    def phi_i(self, n: int, r: float) -> complex:
        """:math:`\\varphi_n^i(r)`

        (Eq. 28, 30)

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.A_in(n) * Bf.besselj(n, self.k_f * r)

    def d_phi(
        self,
        n: int,
        r: float,
        scattered: bool,
        incident: bool,
    ) -> complex:
        """:math:`\\partial_r \\varphi_n(r)`

        (Eq. 28, 30)

        :param n: mode
        :param r: radial coordinate [m]
        :param scattered: scattered field
        :param incident: incident field
        """
        if incident and not scattered:
            return self.d_phi_i(n, r)
        elif incident and scattered:
            return self.d_phi_i(n, r) + self.d_phi_sc(n, r)
        elif not incident and scattered:
            return self.d_phi_sc(n, r)
        else:
            raise ValueError(
                "Neither scattered nor incident field has "
                "been has been selected. Velocity field is zero.",
            )

    def d_phi_sc(self, n: int, r: float) -> complex:
        """:math:`\\partial_r \\varphi_n^{sc}(r)`

        (Eq. 28, 30)

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.a(n) * self.k_f * Bf.d1_hankelh1(n, self.k_f * r)

    def d_phi_i(self, n: int, r: float) -> complex:
        """:math:`\\partial_r \\varphi_n^{i}(r)`

        (Eq. 28, 30)

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.A_in(n) * self.k_f * Bf.d1_besselj(n, self.k_f * r)


if __name__ == "__main__":
    pass



import unittest

import numpy as np

from osaft.core.backgroundfields import WaveType
from osaft.core.functions import sqrt
from osaft.tests.basetest_arf import HelperStandingARF, HelperTravelingARF
from osaft.tests.solution_factory import SolutionFactory
from osaft.tests.solutions.basedoinikov1994.test_base import (
    BaseTestDoinikov1994,
)


class TestARFBase(BaseTestDoinikov1994):
    def setUp(self) -> None:
        super().setUp()

        self.small_boundary_layer = False
        self.large_boundary_layer = False

        self.cls = SolutionFactory().doinikov_1994_compressible_arf()
        self.cls.small_boundary_layer = self.small_boundary_layer
        self.cls.small_boundary_layer = self.small_boundary_layer
        self.cls.large_boundary_layer = self.large_boundary_layer

        self.list_cls = [self.cls]

    def assign_parameters(self) -> None:
        super().assign_parameters()

        self.cls.small_boundary_layer = self.small_boundary_layer
        self.cls.large_boundary_layer = self.large_boundary_layer
        self.cls.N_max = self.N_max

    @property
    def kappa_t(self) -> float:
        return self.rho_f * self.c_f**2 / self.rho_s / self.c_s**2

    @property
    def mu(self) -> float:
        return sqrt(self.rho_f * self.eta_f / self.rho_s / self.eta_s)

    @property
    def g1(self):
        return self.g5 * self.g6

    @property
    def g2(self):
        out = 209 + 148 * self.eta_t + 48 / self.eta_t
        out *= -1 / 9 / self.zeta_t

        out += 19 + 38 * self.eta_t - 12 / self.eta_t

        return out

    @property
    def g3(self):
        out = 25 + 370 * self.eta_t - 80 / self.eta_t
        out *= 1 / 9 / self.zeta_t

        out += -12 - 19 * self.eta_t - 4 / self.eta_t

        return out

    @property
    def g4(self):
        out = 5 + 74 * self.eta_t - 16 / self.eta_t
        out *= 1 / 9 / self.zeta_t

        out += -3 - 4 / self.eta_t

        return out

    @property
    def g5(self):
        out = 1 - self.eta_t
        out *= 19 + 16 / self.eta_t
        return out

    @property
    def g6(self):
        return 89 + 48 / self.eta_t + 38 * self.eta_t

    @property
    def g7(self):
        return self.g6**2

    @property
    def G(self):
        return (self.S1 + self.S2 + self.S3) / 3

    @property
    def S1(self):
        return 1 - self.kappa_t + 0.9 * (1 - 1 / self.rho_t)

    @property
    def S2(self):
        out = 2 * self.f2 - 1
        out *= 2 * self.f1

        out -= 4 * self.f4
        return out

    @property
    def S3(self):
        out = 3 - 1 / self.eta_t
        out *= -120 * (self.f3 - self.f4)

        out += 50 * self.f1 + 43 - 10 * self.kappa_t

        out *= self.rho_s - self.rho_f
        out /= 30 * self.rho_f * (3 + 2 / self.eta_t)

        return out

    @property
    def f1(self):
        return self.g1 / self.g7

    @property
    def f2(self):
        if self.g1 == 0:
            return 0
        return self.g2 * self.g5 / 2 / self.g1

    @property
    def f3(self):
        return self.g3 * self.g6 / self.g7 / 10

    @property
    def f4(self):
        return self.g4 * self.g6 / self.g7 / 2

    @property
    def zeta_t(self):
        return self.zeta_s / self.zeta_f

    @property
    def eta_t(self):
        return self.eta_s / self.eta_f


class TestARFMethods(TestARFBase):
    def test_small_particle_small_delta_standing_not_implemented(self) -> None:
        self.cls.small_boundary_layer = True
        self.cls.small_particle_limit = True
        self.cls.wave_type = WaveType.STANDING
        self.assertRaises(
            NotImplementedError,
            self.cls.compute_arf,
        )

    def test_small_viscous_boundary_layer(self) -> None:
        self.assign_parameters()
        self.assertEqual(
            self.small_boundary_layer,
            self.cls.small_boundary_layer,
        )
        self.small_viscous_boundary_layer = True
        self.assign_parameters()
        self.assertEqual(
            self.small_boundary_layer,
            self.cls.small_boundary_layer,
        )

    def test_small_large_BL_ValueError(self):
        self.parameters.wave_type = WaveType.STANDING
        self.small_boundary_layer = True
        self.large_boundary_layer = True
        self.assign_parameters()

        self.assertRaises(
            ValueError,
            self.cls.compute_arf,
        )

    def test_properties(self) -> None:
        properties = [
            "g1",
            "g2",
            "g3",
            "g4",
            "g5",
            "g6",
            "g7",
            "G",
            "S1",
            "S2",
            "S3",
            "f1",
            "f2",
            "f3",
            "f4",
            "eta_t",
            "zeta_t",
            "rho_t",
            "kappa_t",
            "mu",
        ]
        self._test_properties(properties)


class TestARFLimit1Traveling(HelperTravelingARF, TestARFBase):
    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.small_boundary_layer = True
        self.large_boundary_layer = False

    def compute_arf(self):
        out = 6 * np.pi * self.rho_f * self.x_0**3
        out *= self.cls.field.abs_A_squared
        out *= self.norm_delta * (self.rho_s - self.rho_f) ** 2
        out /= (1 + self.mu) * (2 * self.rho_s + self.rho_f) ** 2

        return out


# |x| << |x_v| << 1
# |x_s | << | x_v_s | << 1
# kappa_t << 1
# rho_f << rho_s


class TestARFLimit2Traveling(HelperTravelingARF, TestARFBase):
    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.small_boundary_layer = False
        self.large_boundary_layer = True

    def compute_arf(self):
        out = np.pi / 6 * self.rho_f * self.x_0**3
        out *= self.norm_delta**2 * self.cls.field.abs_A_squared
        out *= 5 - 2 * self.kappa_t - 8 * self.f1

        return out


class TestARFLimit2Standing(HelperStandingARF, TestARFBase):
    def setUp(self) -> None:
        TestARFBase.setUp(self)
        self.small_boundary_layer = False
        self.large_boundary_layer = True

    def compute_arf(self):
        out = np.pi * self.rho_f * self.x_0**3 * self.cls.field.abs_A_squared
        out *= self.G * np.sin(2 * self.position * self.cls.k_f.real)

        return out


if __name__ == "__main__":
    unittest.main()



from matplotlib import pyplot as plt


class BasePlot:
    """Parent class for all plotting classes."""

    @staticmethod
    def _create_figure(ax: plt.Axes) -> [plt.Figure, plt.Axes]:
        """
        Creates a matplotlib Figure instance and a matplotlib Axes instance.
        If `ax` is passed no new instances are created.

        :param ax: axes
        """
        if ax is None:
            fig, ax = plt.subplots()
            return fig, ax
        else:
            return ax.get_figure(), ax


if __name__ == "__main__":
    pass



import unittest

from osaft.core.functions import BesselFunctions as Bf
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.basetest_scattering import (
    HelperCompareScattering,
    HelperScattering,
)
from osaft.tests.solution_factory import SolutionFactory


class TestScattering(BaseTestSolutions, HelperScattering):
    def setUp(self) -> None:

        BaseTestSolutions.setUp(self)

        self.cls = SolutionFactory().yosioka_1955_scattering()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def A_n(self, n: int) -> complex:

        lamb = self.cls.lambda_rho
        k_f = self.cls.k_f
        k_s = self.cls.k_s
        x_f = self.R_0 * k_f
        x_s = self.R_0 * k_s
        j = lambda x: Bf.besselj(n, x)
        dj = lambda x: Bf.d1_besselj(n, x)
        h = lambda x: Bf.hankelh2(n, x)
        dh = lambda x: Bf.d1_hankelh2(n, x)

        A_num = lamb * k_f * j(x_s) * dj(x_f) - k_s * dj(x_s) * j(x_f)
        A_denom = k_s * dj(x_s) * h(x_f) - lamb * k_f * j(x_s) * dh(x_f)
        A = A_num / A_denom * self.cls.field.A_in(n)

        return A

    def B_n(self, n: int) -> complex:

        lamb = self.cls.lambda_rho
        k_f = self.cls.k_f
        k_s = self.cls.k_s
        x_f = self.cls.x_f
        x_s = self.cls.x_s
        j = lambda x: Bf.besselj(n, x)
        dj = lambda x: Bf.d1_besselj(n, x)
        h = lambda x: Bf.hankelh2(n, x)
        dh = lambda x: Bf.d1_hankelh2(n, x)

        B_num = 1j * k_f
        B_denom = k_s * dj(x_s) * h(x_f)
        B_denom -= lamb * k_f * j(x_s) * dh(x_f)
        B_denom *= x_f**2
        B = B_num / B_denom * self.cls.field.A_in(n)

        return B

    def potential_coefficient(self, n: int) -> complex:
        return -self.A_n(n)

    def V_r_sc(self, n, r):
        A_n = self.cls.A_n(n)
        k_f = self.cls.k_f
        return k_f * A_n * Bf.d1_hankelh2(n, k_f * r)

    def V_theta_sc(self, n, r):
        A_n = self.cls.A_n(n)
        k_f = self.cls.k_f
        return A_n * Bf.hankelh2(n, k_f * r) / r

    def radial_particle_velocity(self, r, theta, t, mode) -> float:
        def radial_func(l, x):
            k = self.cls.k_s
            B = self.cls.B_n(l)
            j = Bf.d1_besselj(l, k * x)
            return B * k * j

        return self.cls.radial_mode_superposition(
            radial_func,
            r,
            theta,
            t,
            mode,
        )

    def tangential_particle_velocity(self, r, theta, t, mode) -> float:
        def radial_func(l, x):
            k = self.cls.k_s
            B = self.cls.B_n(l)
            j = Bf.besselj(l, k * x)
            return B * j / r

        return self.cls.tangential_mode_superposition(
            radial_func,
            r,
            theta,
            t,
            mode,
        )

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_potentials_interface(self) -> None:
        """
        ensures the boundary condition:
        The potential at the particle / fluid - interface is similar.
        """

        def potential_inside(l, x, th, time):
            return self.rho_s * self.cls.Phi_star(x, th, time, l)

        def potential_outside(l, x, th, time):
            return self.rho_f * self.cls.Phi_1(x, th, time, l)

        self.parameters.list_parameters.remove(self.parameters._R_0)
        for n in range(self.n_runs, -1, -1):
            theta = self.get_random_theta()
            t = self.get_random_t()
            self.do_testing(
                func_1=potential_outside,
                args_1=(n, self.R_0, theta, t),
                func_2=potential_inside,
                args_2=(n, self.R_0, theta, t),
            )

    def test_methods_n(self) -> None:
        methods = {}
        methods["A_n"] = None
        methods["B_n"] = None
        methods["potential_coefficient"] = None
        self._test_methods_n(methods, n_end=6, threshold=1e-6)

    def test_V_r_sc(self):
        for n in range(self.n_runs):
            r = self.get_random_r(self.cls)
            self.do_testing(
                func_1=self.cls.V_r_sc,
                args_1=(n, r),
                func_2=self.V_r_sc,
                args_2=(n, r),
            )

    def test_V_theta_sc(self):
        for n in range(self.n_runs):
            r = self.get_random_r(self.cls)
            self.do_testing(
                func_1=self.cls.V_theta_sc,
                args_1=(n, r),
                func_2=self.V_theta_sc,
                args_2=(n, r),
            )


class TestScatteringCompareToKing(BaseTestSolutions, HelperCompareScattering):
    def setUp(self) -> None:

        BaseTestSolutions.setUp(self)

        self._scattering_compare_threshold = 1e-3

        self.parameters.c_s = 1e9 * self.c_f
        self.parameters._c_s.low = 1e9 * self.c_f

        self.cls = SolutionFactory().yosioka_1955_base()

        self.compare_cls = SolutionFactory().king_1934_scattering()

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import numpy as np

from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.warnings import EPS, raise_assumption_warning
from osaft.solutions.doinikov1994compressible.arf_general import (
    ARFArbitraryRadius,
)


class ARF(ARFArbitraryRadius):
    """ARF class for Doinikov (viscous fluid-viscous sphere; 1994)

    There is a large number of options for the computation of the ARF in
    different limiting cases. For a more detailed explanation see
    :ref:`this example <sphx_glr_examples_tutorial_example_doinikov_1994.py>`.

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param c_s: Speed of sound of in the sphere [m/s]
    :param eta_s: shear viscosity of in the sphere  [Pa s]
    :param zeta_s: bulk viscosity of in the sphere [Pa s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity fluid [Pa s]
    :param zeta_f: bulk viscosity fluid [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, traveling or standing
    :param position: Position within the standing wave field [m]
    :param small_boundary_layer: :math:`x \\ll x_v \\ll 1`
    :param large_boundary_layer: :math:`x \\ll 1 \\ll x_v`
    :param N_max: Highest order mode
    :param background_streaming: background streaming contribution
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        c_s: float,
        eta_s: float,
        zeta_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: WaveType = WaveType.STANDING,
        position: None | float = None,
        small_boundary_layer: bool = False,
        large_boundary_layer: bool = False,
        background_streaming: bool = True,
        N_max: int = 5,
    ) -> None:
        """Constructor method"""

        # init of parent class
        super().__init__(
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            c_s=c_s,
            eta_s=eta_s,
            zeta_s=zeta_s,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
            small_boundary_layer=small_boundary_layer,
            large_boundary_layer=large_boundary_layer,
            N_max=N_max,
            background_streaming=background_streaming,
        )

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def compute_arf(self) -> float:
        """Acoustic radiation fore in [N]

        It logs the current values of :attr:`x` and :attr:`x_v`.

        :raises WrongWaveTypeError: if wrong :attr:`wave_type`
        :raises AssumptionWarning: if used solution might not be valid
        """

        # Checking Wave Type
        self.check_wave_type()

        # Checking x_v and x
        abs_x = np.abs(self.x)
        abs_x_v = np.abs(self.x_v)

        # Cases
        if self.small_boundary_layer and (not self.large_boundary_layer):
            test = abs_x < EPS and 1 < EPS * abs_x_v
            raise_assumption_warning(test)
            return self._arf_small_particle_small_delta()
        elif not self.small_boundary_layer and self.large_boundary_layer:
            test = abs_x < EPS * abs_x_v and abs_x_v < EPS
            raise_assumption_warning(test)
            return self._arf_small_particle_large_delta()
        elif not self.small_boundary_layer and not self.large_boundary_layer:
            return self._general_solution()
        else:
            raise ValueError(
                "Small and large viscous boundary layer options cannot"
                "both be True.",
            )


if __name__ == "__main__":
    pass



from __future__ import annotations

from collections.abc import Sequence

from osaft import log
from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.geometries import Sphere
from osaft.core.helper import InputHandler
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveListVariable
from osaft.solutions.base_scattering import BaseScattering
from osaft.solutions.hasegawa1969.base import BaseHasegawa


class ScatteringField(BaseHasegawa, BaseScattering):
    """Scattering field class for Hasegawa (1969)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the solid elastic sphere [kg/m^3]
    :param E_s: Young's modulus [Pa]
    :param nu_s: Poisson's ratio of the solid sphere [-]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: progressive wave
    :param position: Position in the standing wave field [rad]
    :param N_max: Highest order mode included in the computation [-]
    """

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float,
        rho_s: float,
        nu_s: float,
        E_s: float,
        rho_f: float,
        c_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
        N_max: int = 5,
    ) -> None:
        """Constructor method"""
        BaseHasegawa.__init__(
            self,
            f,
            R_0,
            rho_s,
            E_s,
            nu_s,
            rho_f,
            c_f,
            p_0,
            wave_type,
            position,
        )
        BaseScattering.__init__(self, N_max)

        log.info("Create ScatteringField")

        # Dependent Variables
        self._c_n = ActiveListVariable(
            self._compute_c_n,
            "Coefficient c_n" " for scattered potential",
        )

        self._c_n_A = ActiveListVariable(
            self._compute_c_n_A,
            "Coefficient c_n" " for scattered potential",
        )

        self._a_n = ActiveListVariable(
            self._compute_a_n,
            "Coefficient a_n" " for potential inside particle",
        )

        self._b_n = ActiveListVariable(
            self._compute_b_n,
            "Coefficient b_n" " for potential inside particle",
        )

        # Dependencies
        self._c_n.is_computed_by(
            self._lambda_rho,
            self.fluid._k_f,
            self.scatterer._k_l,
            self.scatterer._k_t,
            self.sphere._R_0,
            self.scatterer._nu_s,
        )
        self._c_n_A.is_computed_by(
            self.field._A_in,
            self._c_n,
        )
        self._a_n.is_computed_by(
            self.fluid._k_f,
            self.scatterer._k_l,
            self.scatterer._k_t,
            self.sphere._R_0,
            self.field._A_in,
            self._c_n,
        )
        self._b_n.is_computed_by(
            self.fluid._k_f,
            self.scatterer._k_l,
            self.scatterer._k_t,
            self.sphere._R_0,
            self.field._A_in,
            self._c_n,
        )

        if type(self) is ScatteringField:
            log.info(str(self))
            log.debug(repr(self))

    def __repr__(self):
        return (
            f"{type(self)}(f={self.f}, R_0={self.R_0}, "
            f"rho_s={self.rho_s}, rho_f={self.rho_f}, "
            f"c_f={self.c_f}, E_s={self.E_s}, nu_s={self.nu_s}"
            f"p_0={self.p_0}, position={self.position}, "
            f"wave_type={self.wave_type}, N_max={self.N_max})"
        )

    def __str__(self):
        out = "Hasegawa's solution to the Scattering field"
        out += " with following properties: \n"
        out += SF.get_str_text("Wave type", "", self.wave_type, "")
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "Pa",
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_f",
            self.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_f",
            self.kappa_f,
            "1/Pa",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Young's modulus",
            "E_s",
            self.E_s,
            "Pa",
        )
        out += SF.get_str_text(
            "Poisson's ratio",
            "nu_s",
            self.nu_s,
            "Pa",
        )
        out += "Other\n"
        out += SF.get_str_text(
            "#modes",
            "N_max",
            self.N_max,
            None,
        )
        return out

    # -------------------------------------------------------------------------
    # User-facing functions
    # -------------------------------------------------------------------------

    def radial_particle_velocity(
        self,
        r: float | Sequence[float],
        theta: float | Sequence[float],
        t: float | Sequence[float],
        mode: None | int = None,
    ) -> float | Sequence[float]:
        """Radial particle velocity [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        def radial_func(l, x):
            term1 = self.k_s_l * self.a_n(l) * Bf.d1_besselj(l, self.k_s_l * x)
            term2 = (
                l * (l + 1) / x * self.b_n(l) * Bf.besselj(l, self.k_s_t * x)
            )
            return term1 - term2

        r, theta, t = InputHandler.handle_input(
            r,
            theta,
            t,
            self.R_0,
            inside_sphere=True,
        )
        out = self.radial_mode_superposition(radial_func, r, theta, t, mode)
        return out

    def tangential_particle_velocity(
        self,
        r: float | Sequence[float],
        theta: float | Sequence[float],
        t: float | Sequence[float],
        mode: None | int = None,
    ) -> float | Sequence[float]:
        """Tangential particle velocity [m/s]

        :param r: radial coordinate [m]
        :param theta: tangential coordinate [rad]
        :param t: time [s]
        :param mode: mode
        """

        def radial_func(l, x):
            term1 = self.a_n(l) * Bf.besselj(l, self.k_s_l * x)
            term2 = self.b_n(l) * (
                Bf.besselj(l, self.k_s_t * x)
                + self.k_s_t * x * Bf.d1_besselj(l, self.k_s_t * x)
            )
            return (term1 - term2) / x

        r, theta, t = InputHandler.handle_input(
            r,
            theta,
            t,
            self.R_0,
            inside_sphere=True,
        )
        out = self.tangential_mode_superposition(
            radial_func,
            r,
            theta,
            t,
            mode,
        )
        return out

    # -------------------------------------------------------------------------
    # Velocity Amplitudes and Velocity Potentials
    # -------------------------------------------------------------------------

    def potential_coefficient(self, n: int) -> complex:
        return -self.c_n_A(n)

    def V_r_sc(
        self,
        n: int,
        r: float | Sequence[float],
    ) -> complex | Sequence[complex]:
        """Radial scattering field velocity term of mode `n`
        without Legendre coefficients

        Returns radial scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.c_n_A(n) * self.k_f * Bf.d1_hankelh1(n, self.k_f * r)

    def V_theta_sc(
        self,
        n: int,
        r: float | Sequence[float],
    ) -> complex | Sequence[complex]:
        """Tangential scattering field velocity term of mode n
        without Legendre coefficients

        Returns tangential scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        return self.c_n_A(n) * Bf.hankelh1(n, self.k_f * r) / r

    # -------------------------------------------------------------------------
    # Coefficients
    # -------------------------------------------------------------------------
    def c_n(self, n: int) -> complex:
        """Coefficient :math:`c_n` [-]

        (Eq. 7)

        :param n: mode [-]
        """
        return self._c_n.item(n)

    def c_n_A(self, n: int) -> complex:
        """Coefficient :math:`c_n` [m]

        (Eq. 7)

        :param n: mode [-]
        """
        return self._c_n_A.item(n)

    def a_n(self, n: int) -> complex:
        """Coefficient :math:`a_n` [m]

        (determined by boundary conditions)

        :param n: mode [-]
        """
        return self._a_n.item(n)

    def b_n(self, n: int) -> complex:
        """Coefficient :math:`b_n` [m]

        (determined by boundary conditions)

        :param n: mode [-]
        """
        return self._b_n.item(n)

    def _compute_c_n(self, n: int) -> complex:
        """Compute c_n according to (8, 9)."""
        term0 = 1 / 2 / self.lambda_rho * self.x_s_t**2
        term1 = self.x_s_l * Bf.d1_besselj(n, self.x_s_l)
        term2 = term1 - Bf.besselj(n, self.x_s_l)
        term3 = 2 * n * (n + 1) * Bf.besselj(n, self.x_s_t)
        term4 = (n + 2) * (n - 1) * Bf.besselj(n, self.x_s_t)
        term4 += self.x_s_t**2 * Bf.d2_besselj(n, self.x_s_t)
        term5 = self.nu_s / (1 - 2 * self.nu_s) * Bf.besselj(n, self.x_s_l)
        term5 -= Bf.d2_besselj(n, self.x_s_l)
        term5 *= self.x_s_l**2
        term6 = -2 * n * (n + 1) * self.x_s_t * Bf.d1_besselj(n, self.x_s_t)
        term6 += term3

        F_n = term0 * ((term1 / term2) - (term3 / term4))
        F_n /= (term5 / term2) - (term6 / term4)

        num = self.x_f * Bf.d1_besselj(n, self.x_f)
        num -= F_n * Bf.besselj(n, self.x_f)
        denom = F_n * Bf.hankelh2(n, self.x_f)
        denom -= self.x_f * Bf.d1_hankelh2(n, self.x_f)

        return num / denom

    def _compute_c_n_A(self, n: int) -> complex:
        """Compute c_n according to (8, 9), multiplied with the incident
        amplitude
        """
        out = self.c_n(n)
        out *= self.A_in(n)
        return out

    def _compute_a_n(self, n: int) -> complex:
        """Compute scattering coefficient a_n"""

        j = Bf.besselj
        dj = Bf.d1_besselj
        ddj = Bf.d2_besselj
        dh = Bf.d1_hankelh1
        cn = self.c_n_A(n)
        An = self.A_in(n)
        x = self.x_f
        x1 = self.x_s_l
        x2 = self.x_s_t

        num = x * ((n**2 + n - 1) * x2 * dj(n, x2) - j(n, x2))
        num *= An * dj(n, x) + cn * dh(n, x)

        denom = (n**2 + n - 1) * x2 * dj(n, x2) - (n**2 + n + 1) * j(n, x2)
        denom *= x1 * dj(n, x1)
        denom += n * (n + 1) * j(n, x2) * (j(n, x1) - x1**2 * ddj(n, x1))

        return num / denom

    def _compute_b_n(self, n: int) -> complex:
        """Compute scattering coefficient b_n"""
        j = Bf.besselj
        dj = Bf.d1_besselj
        ddj = Bf.d2_besselj
        dh = Bf.d1_hankelh1
        cn = self.c_n_A(n)
        An = self.A_in(n)
        x = self.x_f
        x1 = self.x_s_l
        x2 = self.x_s_t

        num = -x * (x1 * (x1 * ddj(n, x1) + dj(n, x1)) - j(n, x1))
        num *= An * dj(n, x) + cn * dh(n, x)

        denom = (n**2 + n + 1) * j(n, x2) - (n**2 + n - 1) * x2 * dj(n, x2)
        denom *= x1 * dj(n, x1)
        denom += n * (n + 1) * j(n, x2) * (x1**2 * ddj(n, x1) - j(n, x1))

        return num / denom


if __name__ == "__main__":
    pass



from abc import abstractmethod

from osaft import WaveType
from osaft.core.backgroundfields import WrongWaveTypeError
from osaft.tests.testing_parameters import RandomParameters


class HelperARF:
    """Baseclass for testing the ARF computation"""

    _threshold = 1e-12

    @abstractmethod
    def compute_arf(self):
        pass

    def test_wrong_wave_type_error(self):
        WaveType.WRONGWAVETYPE = 3, "wrong wave type"
        self.cls.wave_type = WaveType.WRONGWAVETYPE
        self.assertRaises(WrongWaveTypeError, self.cls.compute_arf)

    def test_arf(self) -> None:
        self.do_testing(
            self.compute_arf,
            self.cls.compute_arf,
            threshold=self._threshold,
            zero=1e-30,
        )


class HelperStandingARF(HelperARF):

    parameters: RandomParameters

    def test_arf(self) -> None:
        self.parameters.wave_type = WaveType.STANDING
        self.parameters._wave_type.list_of_values = [WaveType.STANDING]
        HelperARF.test_arf(self)


class HelperTravelingARF(HelperARF):

    parameters: RandomParameters

    def test_arf(self) -> None:

        self.parameters.wave_type = WaveType.TRAVELLING
        self.parameters._wave_type.list_of_values = [WaveType.TRAVELLING]
        HelperARF.test_arf(self)


class HelperCompareARF:
    arf_compare_threshold = 1e-12

    def test_comparison_arf(self):
        self.do_testing(
            self.cls.compute_arf,
            self.compare_cls.compute_arf,
            threshold=self.arf_compare_threshold,
            zero=1e-30,
        )



import unittest

from osaft.solutions.base_scattering import BaseScattering
from osaft.tests.plotting.scattering.setup_test_scattering import (
    BaseTestFluidPotentialPlotter,
    BaseTestFluidPressurePlotter,
    BaseTestFluidVelocityPlotter,
)
from osaft.tests.solution_factory import SolutionFactory


class TestVelocityField(BaseTestFluidVelocityPlotter):
    def setUp(self) -> None:

        super().setUp()
        self.name_prefix = "Velocity_Plot_Yosioka_"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().yosioka_1955_scattering()


class TestPressureField(BaseTestFluidPressurePlotter):
    def setUp(self) -> None:
        super().setUp()
        self.name_prefix = "Pressure_Plot_Yosioka_"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().yosioka_1955_scattering()


class TestPotentialField(BaseTestFluidPotentialPlotter):
    def setUp(self) -> None:
        super().setUp()
        self.name_prefix = "Potential_Plot_Yosioka_"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().yosioka_1955_scattering()


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from osaft.core.functions import pi
from osaft.core.helper import InputHandler as IH


class TestInputHandler(unittest.TestCase):
    def setUp(self):
        self.R_0 = 1
        self.arr_r_inside = np.linspace(0, 1)
        self.arr_r_outside = np.linspace(1, 2)
        self.arr_theta = np.linspace(0, pi)
        self.arr_time = np.linspace(0, 1)

    def test_ValueError_float(self):
        self.assertRaises(
            ValueError,
            IH.handle_input,
            2.1,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )
        self.assertRaises(
            ValueError,
            IH.handle_input,
            2.1,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )
        self.assertRaises(
            ValueError,
            IH.handle_input,
            -2.1,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )
        self.assertRaises(
            ValueError,
            IH.handle_input,
            6.1,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )
        self.assertRaises(
            ValueError,
            IH.handle_input,
            -6.1,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )

    def test_ValueError_list(self):
        lst = [self.R_0 * (1 + np.random.rand()) for _ in np.arange(10)]

        lst[2] = 0.99 * self.R_0
        lst[3] = 1.0 * self.R_0
        self.assertRaises(
            ValueError,
            IH.handle_input,
            lst,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            False,
        )

        lst[2] = -1.01 * self.R_0
        self.assertRaises(
            ValueError,
            IH.handle_input,
            lst,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )
        self.assertRaises(
            ValueError,
            IH.handle_input,
            lst,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            False,
        )

    def test_ValueError_np_ndarray(self):
        self.assertRaises(
            ValueError,
            IH.handle_input,
            self.arr_r_outside,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            True,
        )
        self.assertRaises(
            ValueError,
            IH.handle_input,
            self.arr_r_inside,
            self.arr_theta,
            self.arr_time,
            self.R_0,
            False,
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from osaft.core.functions import exp
from osaft.plotting.datacontainers.scattering_datacontainer import (
    FluidScatteringData,
)
from osaft.tests.solution_factory import SolutionFactory


class TestFluidScatteringDatacontainer(unittest.TestCase):
    def setUp(self) -> None:

        super().setUp()

        self.sol = SolutionFactory().yosioka_1955_scattering()
        self.cls = FluidScatteringData(
            self.sol,
            5 * self.sol.R_0,
        )

    def test_r_min(self):
        self.assertEqual(self.cls.r_min, self.sol.R_0)

    def test_velocity_magnitude(self):
        self.cls.scattered = True
        self.cls.incident = True
        # Instantaneous
        _, _, u_mag = self.cls.get_velocity_magnitude(
            instantaneous=True,
            scattered=True,
            incident=True,
        )
        _, _, u, v = self.cls.get_velocity_vector(scattered=True)
        u_mag_tests = np.hypot(u, v)
        np.testing.assert_allclose(u_mag, u_mag_tests)
        # Average
        _, _, u_mag = self.cls.get_velocity_magnitude(
            instantaneous=False,
            scattered=True,
            incident=True,
        )
        u_mag_tests = np.hypot(np.abs(self.cls.u), np.abs(self.cls.w))
        np.testing.assert_allclose(u_mag, u_mag_tests)
        # Instantaneous With Phase
        phase = 1
        _, _, u_mag = self.cls.get_velocity_magnitude(
            instantaneous=True,
            phase=phase,
            scattered=True,
            incident=True,
        )
        _, _, u, v = self.cls.get_velocity_vector(phase=phase)
        u_mag_tests = np.hypot(u, v)
        np.testing.assert_allclose(u_mag, u_mag_tests)

    def test_compare_velocity_displacement(self):
        _, _, v = self.cls.get_velocity_magnitude()
        _, _, u = self.cls.get_displacement_magnitude()

    def test_displacement_phase(self):
        self.cls.scattered = True
        self.cls.incident = False
        phase = 1
        _, _, u, v = self.cls.get_displacement_vector(
            phase=phase,
            scattered=True,
            incident=False,
        )
        u_test = (self.cls.u * exp(-1j * phase) / self.sol.omega).real
        v_test = (self.cls.w * exp(-1j * phase) / self.sol.omega).real
        np.testing.assert_allclose(u, u_test)
        np.testing.assert_allclose(v, v_test)

    def test_velocity_phase(self):
        phase = 1
        _, _, u, v = self.cls.get_velocity_vector(
            phase=phase,
            scattered=True,
            incident=False,
        )
        u_test = (self.cls.u * exp(-1j * phase)).real
        v_test = (self.cls.w * exp(-1j * phase)).real
        np.testing.assert_allclose(u, u_test)
        np.testing.assert_allclose(v, v_test)

    def test_incident_scattering(self):
        _, _, u_incident, v_incident = self.cls.get_velocity_vector(
            scattered=False,
            incident=True,
        )
        _, _, u_scattered, v_scattered = self.cls.get_velocity_vector(
            scattered=True,
            incident=False,
        )
        _, _, u_total, v_total = self.cls.get_velocity_vector(
            scattered=True,
            incident=True,
        )
        np.testing.assert_allclose(u_incident + u_scattered, u_total)
        np.testing.assert_allclose(v_incident + v_scattered, v_total)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from scipy.special import exp1, factorial, lpmv
from sympy.physics.quantum.cg import CG

from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import LegendreFunctions as Leg
from osaft.core.functions import (
    cartesian_2_spherical_coordinates,
    cartesian_2_spherical_vector,
)
from osaft.core.functions import clebsch_gordan_coefficient as cg
from osaft.core.functions import (
    cos,
    exp,
    full_range,
    integrate,
    integrate_osc,
    pi,
    sin,
    spherical_2_cartesian_coordinates,
    spherical_2_cartesian_vector,
    sqrt,
    xexpexp1,
)
from osaft.tests.basetest_numeric import NumericTestCase


class TestFunctions(NumericTestCase):

    # -------------------------------------------------------------------------
    # Range
    # -------------------------------------------------------------------------

    def test_full_range(self):
        res = full_range(1)
        self.assertEqual(res, range(0, 2))

        res = full_range(-1, 3)
        self.assertEqual(res, range(-1, 4))

        self.assertRaises(ValueError, full_range, 1, 0, 3)

    # -------------------------------------------------------------------------
    # Coordinate Transforms
    # -------------------------------------------------------------------------

    def test_cartesian_2_spherical(self):

        x = 0
        z = 0
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, 0)
        self.assertEqual(theta, 0)

        x = 1
        z = 1
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, sqrt(2))
        self.assertEqual(theta, pi / 4)

        x = -1
        z = 0
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, 1)
        self.assertEqual(theta, -pi / 2)

        x = -1
        z = 1
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, sqrt(2))
        self.assertEqual(theta, -pi / 4)

        x = -1
        z = -1
        r, theta = cartesian_2_spherical_coordinates(x, z)
        self.assertEqual(r, sqrt(2))
        self.assertEqual(theta, -pi * 3 / 4)

    def test_spherical_2_cartesian(self):

        r = 0
        theta = 0
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 0)
        self.assertAlmostEqual(z, 0)

        r = 0
        theta = pi / 2
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 0)
        self.assertAlmostEqual(z, 0)

        r = 2
        theta = pi / 4
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 2 * sqrt(2) / 2)
        self.assertAlmostEqual(z, 2 * sqrt(2) / 2)

        r = 3
        theta = pi / 4
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 3 * sqrt(2) / 2)
        self.assertAlmostEqual(z, 3 * sqrt(2) / 2)

        r = 2
        theta = pi / 2
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 2)
        self.assertAlmostEqual(z, 0, threshold=1e-10)

        r = 2
        theta = pi
        x, z = spherical_2_cartesian_coordinates(r, theta)
        self.assertAlmostEqual(x, 0, threshold=1e-10)
        self.assertAlmostEqual(z, -2)

    def test_spherical_2_cartesian_velocities(self):

        theta = 3 / 7 * pi
        v_r = 10
        v_theta = -1
        v_x = sin(theta) * v_r + cos(theta) * v_theta
        v_z = cos(theta) * v_r - sin(theta) * v_theta
        self.assertEqual(
            (v_x, v_z),
            spherical_2_cartesian_vector(v_r, v_theta, theta),
        )

        theta = 0
        v_r = 10
        v_theta = -1
        v_x = sin(theta) * v_r + cos(theta) * v_theta
        v_z = cos(theta) * v_r - sin(theta) * v_theta
        self.assertEqual(
            (v_x, v_z),
            spherical_2_cartesian_vector(v_r, v_theta, theta),
        )

        theta = -pi
        v_r = 0
        v_theta = -1
        v_x = sin(theta) * v_r + cos(theta) * v_theta
        v_z = cos(theta) * v_r - sin(theta) * v_theta
        self.assertEqual(
            (v_x, v_z),
            spherical_2_cartesian_vector(v_r, v_theta, theta),
        )

    def test_cartesian_2_spherical_velocities(self):

        theta = 0
        v_x = 10
        v_z = 0
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, 0)
        self.assertAlmostEqual(v_theta, v_x)

        theta = pi / 4
        v_x = 10
        v_z = 0
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_x * sqrt(2) / 2)
        self.assertAlmostEqual(v_theta, v_x * sqrt(2) / 2)

        theta = pi / 2
        v_x = 10
        v_z = 0
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_x)
        self.assertAlmostEqual(v_theta, 0)

        theta = 0
        v_x = 0
        v_z = 10
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_z)
        self.assertAlmostEqual(v_theta, 0)

        theta = pi / 4
        v_x = 0
        v_z = 10
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, v_z * sqrt(2) / 2)
        self.assertAlmostEqual(v_theta, -v_z * sqrt(2) / 2)

        theta = pi / 2
        v_x = 0
        v_z = 10
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, 0)
        self.assertAlmostEqual(v_theta, -v_z)

        theta = 0
        v_x = 10
        v_z = 10
        speed = np.hypot(v_x, v_z)
        v_r, v_theta = cartesian_2_spherical_vector(v_x, v_z, theta)
        self.assertAlmostEqual(v_r, speed * sqrt(2) / 2)
        self.assertAlmostEqual(v_theta, speed * sqrt(2) / 2)

    # -------------------------------------------------------------------------
    # Integration
    # -------------------------------------------------------------------------

    def test_integrate(self):
        def func(x):
            return sin(x) ** 2

        self.assertAlmostEqual(integrate(func, 0, 2 * pi)[0], pi)
        self.assertAlmostEqual(integrate(func, 0, 2 * pi, 1e-12)[0], pi)

        def func(x):
            return -2 * x + 1

        self.assertAlmostEqual(integrate(func, 0, 2)[0], -2)
        self.assertAlmostEqual(integrate(func, -2, 1 / 2)[0], 6.25)
        self.assertAlmostEqual(integrate(func, -2, 1 / 2, 1e-12)[0], 6.25)

    def test_integrate_osc(self):
        def h1_0(z):
            return Bf.hankelh1(0, k_v * z).real

        def h1_1(z):
            return Bf.hankelh1(1, k_v * z).real

        def h1_2(z):
            return Bf.hankelh1(2, k_v * z).real

        def h1_3(z):
            return Bf.hankelh1(3, k_v * z).real

        def j1_0(z):
            return Bf.besselj(0, k_v * z).real

        def j1_1(z):
            return Bf.besselj(1, k_v * z).real

        def j1_2(z):
            return Bf.besselj(2, k_v * z).real

        def j1_3(z):
            return Bf.besselj(3, k_v * z).real

        # Parameters
        R0 = 8e-6
        lambda_v = 30e-6
        delta = 50e-6
        x1 = R0
        x2 = R0 + 20 * delta
        k_v = 2 * pi / lambda_v + 1j / delta

        # Zeroth Order Bessel and Hankel
        self.assertAlmostEqual(
            integrate_osc(h1_0, x1, x2, delta, lambda_v)[0],
            8.37973712155099e-7,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_1, x1, x2, delta, lambda_v)[0],
            2.397545981955677e-6,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_2, x1, x2, delta, lambda_v)[0],
            2.219617723321184e-6,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_3, x1, x2, delta, lambda_v)[0],
            4.026448321970975e-7,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_0, x1, x2, delta, lambda_v)[0],
            4.38223531947702,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_1, x1, x2, delta, lambda_v)[0],
            4.659233107792744,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_2, x1, x2, delta, lambda_v)[0],
            -4.310991826110144,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_3, x1, x2, delta, lambda_v)[0],
            -4.750479141883805,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_0, x1, x2, delta)[0],
            8.37973712155099e-7,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_1, x1, x2, delta)[0],
            2.397545981955677e-6,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_2, x1, x2, delta)[0],
            2.219617723321184e-6,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(h1_3, x1, x2, delta)[0],
            4.026448321970975e-7,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_0, x1, x2, delta)[0],
            4.38223531947702,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_1, x1, x2, delta)[0],
            4.659233107792744,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_2, x1, x2, delta)[0],
            -4.310991826110144,
            threshold=1e-7,
        )
        self.assertAlmostEqual(
            integrate_osc(j1_3, x1, x2, delta)[0],
            -4.750479141883805,
            threshold=1e-7,
        )

    # -------------------------------------------------------------------------
    # Clebsch Gordan
    # -------------------------------------------------------------------------

    def test_clebsch_gordan(self):
        def random_value():
            return np.random.randint(0, 4)

        nbr_nonzero = 0

        # Test function for int values

        while nbr_nonzero < 20:
            j1 = random_value()
            j2 = random_value()
            m1 = random_value()
            m2 = random_value()
            j = random_value()
            m = random_value()

            result = cg(j1, m1, j2, m2, j, m)

            self.assertAlmostEqual(
                result,
                float(CG(j1, m1, j2, m2, j, m).doit()),
            )

            if result != 0:
                nbr_nonzero += 1

        # Test errors for non int inputs
        numbers = [1.5, 1, 1, 1, 1, 1]

        for i in np.arange(6):
            j = numbers[i % 6]
            j1 = numbers[(i + 1) % 6]
            j2 = numbers[(i + 2) % 6]
            m = numbers[(i + 3) % 6]
            m1 = numbers[(i + 4) % 6]
            m2 = numbers[(i + 5) % 6]

        self.assertRaises(ValueError, cg, j1, m1, j2, m2, j, m)

    # -------------------------------------------------------------------------
    # Legendre Polynomials
    # -------------------------------------------------------------------------

    def test_legendre(self):

        # first_legendre_cos_poly
        def tmp(theta, coefficients):
            out = 0
            for n, c in enumerate(coefficients):
                out += c * lpmv(1, n, cos(theta))
            return out

        for _ in np.arange(10):
            theta = 2 * pi * np.random.rand()
            coefficients = np.random.rand(10) + 1j * np.random.rand(10)
            n = np.random.randint(0, 5)

            # Test Monomial
            # legendre_cos_monomial
            self.assertEqual(
                Leg.cos_monomial(n, theta, coefficients[0]),
                coefficients[0] * lpmv(0, n, cos(theta)),
            )

            # first_legendre_cos_monomial
            self.assertEqual(
                Leg.first_cos_monomial(n, theta, coefficients[0]),
                coefficients[0] * lpmv(1, n, cos(theta)),
            )

            # Test Polynomials
            theta = 2 * pi * np.random.rand(5)
            for index, th in enumerate(theta):
                # legendre_cos_poly
                self.assertEqual(
                    Leg.cos_poly(theta, coefficients)[index],
                    np.polynomial.legendre.legval(
                        cos(th),
                        coefficients,
                    ),
                )
                # legendre_first_cos_poly
                self.assertAlmostEqual(
                    Leg.first_cos_poly(theta, coefficients)[index],
                    tmp(th, coefficients),
                )

    # -------------------------------------------------------------------------
    # Exponential Integral
    # -------------------------------------------------------------------------

    def test_xexpexp1(self):
        def compare_function(z: complex, N: int = 20):
            if z.real > 20:
                out = 0
                for n in full_range(N):
                    out += factorial(n) / (-z) ** n
                return out
            else:
                return z * exp(z) * exp1(z)

        for x in np.linspace(1, 100, 10):
            z1 = x * (1 + 1j)
            z2 = -1j * z1
            z3 = z1 + z2
            self.assertAlmostEqual(xexpexp1(z1), compare_function(z1))
            self.assertAlmostEqual(xexpexp1(z2), compare_function(z2))
            self.assertAlmostEqual(xexpexp1(z3), compare_function(z3))

    def _assert_equal_complex(self, z1, z2):
        threshold = 1e-8 * np.abs(z2[0])

        self.assertTrue(np.abs(np.real(z1[0]) - np.real(z2[0])) < threshold)
        self.assertTrue(np.abs(np.imag(z1[0]) - np.imag(z2[0])) < threshold)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from osaft import log
from osaft.core.functions import pi
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.doinikov2021viscous.streaming import StreamingField


class ARF(StreamingField, BaseARF):
    """ARF class for Doinikov (viscous fluid-elastic sphere; 2021)

    :param f: frequency [Hz]
    :param R_0: radius [m]
    :param rho_s: density of the particle [kg/m^3]
    :param E_s: Young's modulus of the particle [Pa]
    :param nu_s: Poisson's ratio of the particle [-]
    :param rho_f: density of the fluid [kg/m^3]
    :param c_f: speed of sound in the fluid [m/s]
    :param eta_f: fluid component shear viscosity [Pa s]
    :param zeta_f: fluid component bulk viscosity [Pa s]
    :param p_0: pressure amplitude [Pa]
    :param wave_type: wave type
    :param position: position in the standing wave field [m]
    :param N_max: truncation of the summation index (highest order mode)
    :param inf_factor: see :attr:`.inf_factor`
    :param inf_type: see :attr:`.inf_type`
    :param integration_rel_eps:  relative eps for adaptive integration
    """

    # -------------------------------------------------------------------------
    # User-facing methods
    # -------------------------------------------------------------------------

    def compute_arf(
        self,
        return_error: bool = False,
    ) -> float | tuple[float, float]:

        self.check_wave_type()

        # Integral 1
        integral_1 = self.I_1()
        error_1 = self.I_1_error()

        # Error estimation
        rel_error = error_1 / integral_1

        if abs(rel_error) > 0.01:  # pragma: no cover
            log.warning(
                f"Estimated relative error is {rel_error} > 0.01 \n"
                f"If you want to decrease it try to change the integration "
                f"precision `int_eps` or the integration boundary "
                f"`inf_factor` and `inf_type` \n",
            )
        else:
            log.info(
                f"Estimated relative error is {rel_error} < 0.01 \n",
            )
        if return_error:
            return integral_1, error_1
        else:
            return integral_1

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_C_0(self, L_max: None | int = 1) -> dict:
        return super()._compute_C_0(L_max=L_max)

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def I_1(self) -> float:
        """Integral :math:`I_1`

        (Eq. 27)
        """

        def term_second_sum(l: int) -> float:

            if l == 0:
                t_1 = 0
            else:
                t_1 = self.V_r(l, self.R_0, True, True)
                t_1 += (l + 1) * self.V_theta(l, self.R_0, True, True)
                t_1 *= self.V_r(l - 1, self.R_0, True, True).conjugate()
                t_1 *= l / ((2 * l - 1) * (2 * l + 1))

            t_2 = self.V_r(l, self.R_0, True, True)
            t_2 -= l * self.V_theta(l, self.R_0, True, True)
            t_2 *= self.V_r(l + 1, self.R_0, True, True).conjugate()
            t_2 *= (l + 1) / ((2 * l + 1) * (2 * l + 3))

            return (t_1 + t_2).real

        term_1 = 4 * self.eta_f / self.R_0**2 * self.C_0[3, 1]
        term_2 = 2 / 3 * self.R_0 * self.rho_f * self.gamma(1, self.R_0, False)
        first_sum = term_1 + term_2

        second_sum = 0
        for n in self.range_N_max:
            second_sum += term_second_sum(n)
        return 2 * pi * self.R_0**2 * (first_sum - self.rho_f * second_sum)

    def I_1_error(self) -> float:
        """Estimated numerical error for integral :math:`I_1`

        (Eq. 27)
        """
        return 8 * self.eta_f * self.C_0_error[3, 1]


if __name__ == "__main__":
    pass



from __future__ import annotations

from abc import ABC, abstractmethod

from osaft.core.functions import BesselFunctions as Bf
from osaft.solutions.base_scattering import BaseScattering


class BaseScatteringDoinikov1994(BaseScattering, ABC):
    """Base scattering class for Doinikov solutions from 1994

    :param N_max: Highest order mode included in the computation [-]
    """

    def __init__(self, N_max: int) -> None:
        """ "init method"""
        super().__init__(N_max)

    def potential_coefficient(self, n: int) -> complex:
        return self.alpha_n(n) * self.field.A_in(n)

    # -------------------------------------------------------------------------
    # Velocity amplitudes
    # -------------------------------------------------------------------------

    def V_r_sc(self, n: int, r: float) -> complex:
        """Radial scattering field velocity term of mode `n`
        without Legendre coefficients

        Returns radial scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        out = self.k_f * self.alpha_n(n) * Bf.d1_hankelh1(n, r * self.k_f)
        out -= n * (n + 1) / r * self.beta_n(n) * Bf.hankelh1(n, r * self.k_v)
        out *= self.A_in(n)
        return out

    def V_theta_sc(self, n: int, r: float) -> complex:
        """Tangential scattering field velocity term of mode n
        without Legendre coefficients

        Returns tangential scattering field velocity in [m/s]

        :param n: mode
        :param r: radial coordinate [m]
        """
        arg = self.k_v * r
        out = Bf.hankelh1(n, arg)
        out += arg * Bf.d1_hankelh1(n, arg)
        out *= -self.beta_n(n)
        out += self.alpha_n(n) * Bf.hankelh1(n, self.k_f * r)
        out *= self.A_in(n)
        return out / r

    # -------------------------------------------------------------------------
    # Abstract methods
    # -------------------------------------------------------------------------

    @property
    @abstractmethod
    def k_f(self) -> complex:
        """Returns the wave number in the fluid :math:`k_f` [1/m]"""
        pass

    @property
    @abstractmethod
    def k_v(self) -> complex:
        """Returns the viscous wave number in the fluid :math:`k_v` [1/m]"""
        pass

    @abstractmethod
    def A_in(self, n: int):
        """Incoming wave amplitude

        :param n: mode
        """
        pass

    @abstractmethod
    def alpha_n(self, n: int) -> complex:
        """:math:`\\alpha_n` coefficient

        :param n: mode
        """
        pass

    @abstractmethod
    def beta_n(self, n: int) -> complex:
        """:math:`\\beta_n` coefficient

        :param n: mode
        """
        pass


if __name__ == "__main__":
    pass
